package demos.calc;

import java.awt.Component;
import java.awt.Toolkit;
import java.awt.datatransfer.*;
import java.awt.event.ActionEvent;
import java.io.IOException;
import java.math.BigDecimal;
import java.math.BigInteger;
import java.util.*;

import static java.math.BigDecimal.ZERO;
import static java.math.BigDecimal.ONE;
import static java.math.BigDecimal.TEN;

import javax.swing.JButton;

import util.Builders;

public final class Calc implements Controller {

    private final CalcGUI gui = new CalcGUI(this);
    private BigDecimal x = ZERO, leftOperand = null;
    private BinOp curOp = null;
    private boolean afterPoint = false;
    private State state = State.LEFT_OPERAND;
    
    private void setX(BigDecimal x) {
        this.x = x;
        updateReadout();
    }
    
    private enum State {
        LEFT_OPERAND, RIGHT_OPERAND, AFTER_EQ
    }
    
    private enum BinOp {
        PLUS {
            protected void exec(Calc self, BigDecimal left, BigDecimal right) {
                self.setX(left.add(right));
            }
        },
        MINUS {
            protected void exec(Calc self, BigDecimal left, BigDecimal right) {
                self.setX(left.subtract(right));
            }
        },
        TIMES {
            protected void exec(Calc self, BigDecimal left, BigDecimal right) {
                self.setX(left.multiply(right));
            }
        },
        DIVIDED {
            protected void exec(Calc self, BigDecimal left, BigDecimal right) {
                self.setX(left.divide(right));
            }
        };
        protected abstract void exec(
                Calc self, BigDecimal left, BigDecimal right);
    }
    
    private void updateReadout() {
        switch (state) {
            case LEFT_OPERAND:
            case RIGHT_OPERAND:
                if (afterPoint)
                    if (isIntegral(x))
                        setReadoutText(x.toPlainString() + ".");
                    else
                        setReadoutText(x.toPlainString());
                else
                    setReadoutText(integerPart(x).toPlainString());
                break;
            case AFTER_EQ:
            default:
                if (isIntegral(x))
                    setReadoutText(integerPart(x).toPlainString());
                else
                    setReadoutText(x.toPlainString());
        }
    }
    
    private void setReadoutText(String text) {
        gui.setTextFieldText(gui.readout, text);
    }
    
    public void quit(Component widget, ActionEvent event) {
        gui.mainWindow.dispose();
    }

    public void copy(Component widget, ActionEvent event) {
        final Clipboard clipboard = 
            Toolkit.getDefaultToolkit().getSystemClipboard();
        final Transferable transferable = new Transferable() {
            public Object getTransferData(DataFlavor flavor)
                    throws UnsupportedFlavorException {
                return x.toString();
            }
        
            public boolean isDataFlavorSupported(DataFlavor flavor) {
                return flavor.equals(DataFlavor.stringFlavor);
            }
            
            private final DataFlavor[] flavors = { DataFlavor.stringFlavor };
        
            public DataFlavor[] getTransferDataFlavors() {
                return flavors;
            }
        };
        final ClipboardOwner clipboardOwner = new ClipboardOwner() {
            public void lostOwnership(
                    Clipboard clipboard, Transferable contents) {}
        };
        
        clipboard.setContents(transferable, clipboardOwner);
    }

    public void paste(Component widget, ActionEvent event) {
        final Clipboard clipboard = 
            Toolkit.getDefaultToolkit().getSystemClipboard();
        
        final String contents;
        try {
            contents = (String) clipboard.getContents(null)
                    .getTransferData(DataFlavor.stringFlavor);
        } catch (UnsupportedFlavorException e) {
            return;
        } catch (IOException e) {
            return;
        }
        
        final BigDecimal decValue;
        try {
            decValue = new BigDecimal(contents);
        } catch (NumberFormatException e) {
            return;
        }
        
        setX(decValue);
    }

    public void help(Component widget, ActionEvent event) {
        
    }

    public void about(Component widget, ActionEvent event) {
        
    }

    public void clear(Component widget, ActionEvent event) {
        if (x.equals(0.0) && !afterPoint) {
            // We already cleared, so reset this time
            curOp = null;
            leftOperand = null;
            state = State.LEFT_OPERAND;
        } else {
            afterPoint = false;
            setX(ZERO);
        }
    }

    private void clearSilently() {
        afterPoint = false;
        x = ZERO;
    }
    
    private final IdentityHashMap<JButton, Integer> digitButtonMap = 
        Builders.buildMap(new IdentityHashMap<JButton, Integer>(10))
            .put(gui.btn0, 0)
            .put(gui.btn1, 1)
            .put(gui.btn2, 2)
            .put(gui.btn3, 3)
            .put(gui.btn4, 4)
            .put(gui.btn5, 5)
            .put(gui.btn6, 6)
            .put(gui.btn7, 7)
            .put(gui.btn8, 8)
            .put(gui.btn9, 9)
            .finish();
    
    public void digit(Component widget, ActionEvent event) {
        if (!(widget instanceof JButton))
            return;
        final JButton btn = (JButton) widget;
        final BigInteger digit = BigInteger.valueOf(digitButtonMap.get(btn));
        
        if (state == State.AFTER_EQ) {
            clear(widget, event);
            state = State.LEFT_OPERAND;
        }
        
        if (afterPoint) {
            BigInteger mantissa = x.unscaledValue();
            int exponent = x.scale();
            
            mantissa = mantissa.multiply(BigInteger.TEN).add(digit);
            setX(new BigDecimal(mantissa, exponent + 1));
        } else
            setX(x.multiply(TEN).add(new BigDecimal(digit)));
    }

    public void point(Component widget, ActionEvent event) {
        afterPoint = true;
        updateReadout();
    }
    
    private void doBinOp(BinOp op) {
        if (state == State.RIGHT_OPERAND) {
            assert leftOperand != null;
            assert curOp != null;
            
            curOp.exec(this, leftOperand, x);
        }
        leftOperand = x;
        curOp = op;
        clearSilently();
        state = State.RIGHT_OPERAND;
    }
    
    public void plus(Component widget, ActionEvent event) {
        doBinOp(BinOp.PLUS);
    }

    public void minus(Component widget, ActionEvent event) {
        doBinOp(BinOp.MINUS);
    }

    public void times(Component widget, ActionEvent event) {
        doBinOp(BinOp.TIMES);
    }

    public void div(Component widget, ActionEvent event) {
        doBinOp(BinOp.DIVIDED);
    }

    public void sqrt(Component widget, ActionEvent event) {
        setX(BigDecimal.valueOf(Math.sqrt(x.doubleValue())));
    }

    public void pct(Component widget, ActionEvent event) {
        setX(x.divide(BigDecimal.valueOf(100)));
        eq(widget, event);
    }

    public void eq(Component widget, ActionEvent event) {
        if (state != State.RIGHT_OPERAND)
            return;
        state = State.AFTER_EQ;
        curOp.exec(this, leftOperand, x);
        curOp = null;
        leftOperand = null;
    }

    public static void main(String ... args) {
        new Calc();
    }

    private static boolean isIntegral(BigDecimal x) {
        return integerPart(x).compareTo(x) == 0;
    }
    
    private static BigDecimal integerPart(BigDecimal x) {
        return x.divideToIntegralValue(ONE);
    }

}

