package errors;

import java.util.*;

import types.Widget;
import util.Functional;

/** An invalid attribute value.
 *  This will likely be thrown only by library code. As such, it permits an
 *  extra error message to be passed along, which is expected to describe the
 *  problem - for example, it might be "width must be an integer". */
public class BadAttributeValueException extends BadAttributeException {
    /** The value of the attribute. */
    public final String attrValue;

    /**
     * @param widget The widget with the invalid attribute.
     * @param attrName The name of the invalid attribute.
     * @param attrValue The invalid value in question.
     */
    public BadAttributeValueException(Widget widget, String attrName,
            String attrValue) {
        super(formatMsg(widget, attrName, attrValue), widget, attrName);
        this.attrValue = attrValue;
    }

    /**
     * @param msg The details to be appended to the error message.
     * @param widget The widget with the invalid attribute.
     * @param attrName The name of the invalid attribute.
     * @param attrValue The invalid value in question.
     */
    public BadAttributeValueException(String msg, Widget widget,
            String attrName, String attrValue) {
        super(formatMsg(msg, widget, attrName, attrValue), widget, attrName);
        this.attrValue = attrValue;
    }
    
    /**
     * @param widget The widget with the invalid attribute.
     * @param attrName The name of the invalid attribute.
     * @param attrValues The whole list of values of the attribute.
     */
    public BadAttributeValueException(Widget widget, String attrName,
            Collection<String> attrValues) {
        super(formatMsg(widget, attrName, 
                Functional.quoteAndJoin(attrValues, ", ")), widget, attrName);
        this.attrValue = Functional.quoteAndJoin(attrValues, ", ");
    }
    /**
     * @param msg The details to be appended to the error message.
     * @param widget The widget with the invalid attribute.
     * @param attrName The name of the invalid attribute.
     * @param attrValues The whole list of values of the attribute.
     */
    public BadAttributeValueException(String msg, Widget widget,
            String attrName, Collection<String> attrValues) {
        super(formatMsg(msg, widget, attrName, 
                Functional.quoteAndJoin(attrValues, ", ")), widget, attrName);
        this.attrValue = Functional.quoteAndJoin(attrValues, ", ");
    }

    private static String formatMsg(Widget widget, String attrName,
            String attrValue) {
        return String.format(
                "%s: invalid value for attribute %s of %s widgets",
                attrValue, attrName, widget.type);
    }

    private static String formatMsg(String msg, Widget widget,
            String attrName, String attrValue) {
        return formatMsg(widget, attrName, attrValue) + ": " + msg;
    }
}
