package errors;

/** The superclass of all exceptions arising from errors in the .gui file.
 *
 *  All constructors of this class require a line number. Pass -1 for an
 *  unknown line number (then fix the code so the line number is known!). */
public abstract class GUIFileException extends Exception {
    /** The number of the line at which the error occurred.
     *  If the error occurred over a range of lines, this should be the first
     *  of those lines. */
    public final int lineNum;

    /** Use the generic exception message.
     *  @param lineNum The line number of the error. */
    public GUIFileException(int lineNum) {
        super();
        this.lineNum = lineNum;
    }

    /** Use a given message.
     *  Prepends "line ##: " to the message, where ## is the line number.
     *  @param msg The error message.
     *  @param lineNum The line number of the error. */
    public GUIFileException(String msg, int lineNum) {
        super(prependLineNum(msg, lineNum));
        this.lineNum = lineNum;
    }

    /** Use a given message and cause.
     *  @param msg The error message.
     *  @param cause The <code>Throwable</code> next in the cause chain.
     *  @param lineNum The line number of the error. */
    public GUIFileException(String msg, Throwable cause, int lineNum) {
        super(prependLineNum(msg, lineNum), cause);
        this.lineNum = lineNum;
    }

    /** Use a given cause and the generic exception message.
     *  @param cause The <code>Throwable</code> next in the cause chain.
     *  @param lineNum The line number of the error. */
    public GUIFileException(Throwable cause, int lineNum) {
        this(null, cause, lineNum);
    }
    
    private static String prependLineNum(String msg, int lineNum) {
        return String.format("line %d: %s", lineNum, msg);
    }
}
