package interpreter;

import interpreter.CommandLineOptions.State;

import java.io.*;
import java.util.*;

import parser.Parser;
import types.ErrorHandler;
import types.GUIFile;
import types.LangDef;
import types.GUIFile.Flag;
import util.CommandLineParser;
import util.CommandLineParser.InvalidCommandLineException;

/**
 *     Assembler.java is the master-class of the interpreter structure.
 * It instantiates and manipulates the parser, instantiates the LangDef
 * and passes it the Parser's results, then finally outputs the LangDef
 * code with the appropriate file names.
 */

public class Assembler {

	/**
     * Run the Elvis compiler from the command line.
	 * @param args Run the program with no arguments for usage information.
	 */
	public static void main(String ... args) throws Throwable {
		final String[][] outputCodeFiles;
        
        final CommandLineParser<CommandLineOptions, State> cmdParser =
            CommandLineParser.buildParser(CommandLineOptions.class);
        
        final CommandLineOptions.State optionValues = 
            new CommandLineOptions.State();
        
		try {			
            // This method call will parse the whole command line, changing
            // optionValues as options are read in, and return the arguments
            // (that is, those arguments that aren't arguments of options).
            final String[] plainArgs = cmdParser.parse(optionValues, args);
            
            if (optionValues.showHelp) {
                System.out.println(cmdParser.getHelp());
                System.exit(1);
            } else if (optionValues.showUsage) {
                System.out.println(cmdParser.getUsage());
                System.exit(1);
            }
            
            if (plainArgs.length < 2) {
                System.out.println(cmdParser.getUsage());
				System.exit(1);
            }
				
            final String inputFileName = plainArgs[0];
            final String languageName = plainArgs[1];
            
			final FileInputStream elvisCodeFile = 
                new FileInputStream(new File(inputFileName));
			
            final ErrorHandler errorHandler = new ErrorHandler.StdErr();
			final GUIFile guiFile = Parser.parse(elvisCodeFile, inputFileName,
                    errorHandler);
			
            if (guiFile == null)
                System.exit(1);
            
            // Look through the options and apply any command-line settings
            if (optionValues.suppressModifiers)
                guiFile.flags.add(Flag.SUPPRESS_MODIFIERS);
            
            if (optionValues.generateMain)
                guiFile.flags.add(Flag.GENERATE_MAIN);
            
            if (optionValues.projectName != null)
                guiFile.setProjectName(optionValues.projectName);
            
            if (optionValues.packageId != null)
                guiFile.setPackageId(optionValues.packageId);
            
            if (optionValues.interactive)
                processOverride(guiFile);
            
            LangDef interpreter = LangDef.defineLanguage(languageName);
            
            outputCodeFiles = 
                interpreter.generateCode(guiFile, errorHandler);
            
            if (errorHandler.anyErrors())
                System.exit(1);
            
            for (int i=0; i<outputCodeFiles.length; i++) {
                FileOutputStream out = new FileOutputStream(outputCodeFiles[i][0]);
                PrintStream pstream = new PrintStream(out); // pstream ... heehee!
                pstream.print(outputCodeFiles[i][1]);
                pstream.close();
                out.close();
            }
        } catch (InvalidCommandLineException e) {
            System.err.println(e.getMessage() + '\n');
            System.err.println(cmdParser.getHelp());
            System.exit(1);
        }
	}
	
	/**
	 * 		generateCode activates the assembler from an external class, outputting the
	 * generated code into their files and then returning the string corresponding to
	 * the main generated class.
	 *  
	 * @param errorHandler
	 * @param inputFileName
	 * @param languageName
	 * @param executable
	 * @param suppressModifiers
	 * @return outputCodeName - the name of the file associated with the main class
	 * @throws Throwable
	 */
	public String generateCode(ErrorHandler errorHandler, String inputFileName,
					String languageName, boolean executable, boolean suppressModifiers)
					throws Throwable {
		final String[][] outputCodeFiles;
            
		final FileInputStream elvisCodeFile = 
            new FileInputStream(new File(inputFileName));
			
		final GUIFile guiFile = Parser.parse(elvisCodeFile, inputFileName,
            errorHandler);
		
        if (guiFile == null)
            return null;
        
        
            // Look through the options and apply any command-line settings
            if (suppressModifiers)
                guiFile.flags.add(Flag.SUPPRESS_MODIFIERS);
            else
            	guiFile.flags.remove(Flag.SUPPRESS_MODIFIERS);
            
            if (executable)
                guiFile.flags.add(Flag.GENERATE_MAIN);
            else
            	guiFile.flags.remove(Flag.GENERATE_MAIN);
            
            LangDef interpreter = LangDef.defineLanguage(languageName);
            
            outputCodeFiles = 
                interpreter.generateCode(guiFile, errorHandler);
            
            if (!errorHandler.anyErrors())
	            for (int i=0; i<outputCodeFiles.length; i++) {
	                FileOutputStream out = new FileOutputStream(outputCodeFiles[i][0]);
	                PrintStream pstream = new PrintStream(out); // pstream ... heehee!
	                pstream.print(outputCodeFiles[i][1]);
	                pstream.close();
	                out.close();
	            }
            return outputCodeFiles[0][0];
	}
	
	/**
	 * processOverrides queries the user as to what settings they desire
	 * to override from the file specifications.
	 *  
	 * @param guiFile
	 */
	private static void processOverride(GUIFile guiFile) {
		Scanner in = new Scanner(System.in);
		
		/*
		 * Specifies the SUPPRESS_MODIFIERS Flag
		 */
		while(true){
			System.out.println("Generate Modifier Methods?");
			System.out.println("1) Supress Modifiers");
			System.out.println("2) Generate Modifiers");
			System.out.println("3) Default to ELVIS code specifications");
		
			String option = in.next();
			if(option.equals("1")){
				guiFile.flags.add(Flag.SUPPRESS_MODIFIERS);
				break;
			} else if (option.equals("2")) {
				guiFile.flags.remove(Flag.SUPPRESS_MODIFIERS);
				break;
			} else if (option.equals("3")){
				break;
			} else {
				System.out.println("Please choose 1, 2 or 3.");
			}
		}
		
		/*
		 * Specifies the GENERATE_MAIN flag
		 */
		while(true){
			System.out.println("Generate Executable Mockup?");
			System.out.println("1) Executable Mockup");
			System.out.println("2) Constructable Final Code");
			System.out.println("3) Default to ELVIS code specifications");
		
			String option = in.next();
			if(option.equals("1")){
				guiFile.flags.add(Flag.GENERATE_MAIN);
				break;
			} else if (option.equals("2")) {
				guiFile.flags.remove(Flag.GENERATE_MAIN);
				break;
			} else if (option.equals("3")){
				break;
			} else {
				System.out.println("Please choose 1, 2 or 3.");
			}
		}
		
		/*
		 * Overrides the project name, if desired.
		 */
		while(true){
			System.out.println("Override project name? (yes/no)");
			String option = in.next().toLowerCase();
			
			if(option.equals("yes")||option.equals("y"))
			{
				System.out.print("New Project Name: ");
				guiFile.setProjectName(in.next());
				break;
			} else if (option.equals("no")||option.equals("n")){
				break;
			} else {
				System.out.println("Please enter yes / no.");
			}
		}		
		in.close();
	}
}
