package lib.java;

import java.util.*;

import errors.BadAttributeException;
import errors.BadAttributeValueException;

import output.java.CodeBuilder;

import types.GUIAttributes;
import types.Widget;
import types.WidgetDef;

import util.Builders;

/*
 * @EName BarLayout Widget Definition for a grid layout in Java.
 * 
 */

abstract class BarLayout extends WidgetDef {
    abstract protected String getAxisId();
    
    public BarLayout(Widget widget, WidgetDef parent) {
        super(widget, parent);
    }
    
    public final String getLanguage() {
        return "Java";
    }

    public String preamble() {
        return String.format(
                "%s = new BoxLayout(%s, BoxLayout.%s);\n", 
                id, parent.id, getAxisId());
    }
    
    private enum VerticalAlignment {
        top(0f), center(0.5f), bottom(1.0f);
        private final float value;
        VerticalAlignment(float value) {
            this.value = value;
        }
    }
    
    public final Map<String, List<String>> getDefaults() {
        return Builders.buildMultiMap(new TreeMap<String, List<String>>())
            .put("gap", "3")  //EDefaultValue gap 3
            .put("alignVert", "center")  //@EDefaultValue alignVert center
            .finish();
    }
    
    private int gap;
    private float alignVert;
    
    public final String addAttributes(GUIAttributes attrs)
            throws BadAttributeException {
    	//@EName gap Specifies the Gap between Elements in the Layout
        gap = attrs.getInteger("gap");
        try {
        	//@EName alignVert Specifies the vertical alignment of elements in the layout.
            alignVert = attrs.getEnum(
                    "alignVert", VerticalAlignment.class).value;
        } catch (BadAttributeValueException e) {
            alignVert = attrs.getFloat("alignVert"); 
        }
        return "";
    }

    @Override
    public final String addChild(String childType, String childId,
            GUIAttributes attrs) {
        if (isDerivedFrom(childType, Spacer.class))
            return String.format(
                    "SPECIAL$%s$layoutChildren.add(%s);\n" +
                    "SPECIAL$%1$s$layoutConstraints.add(null);\n",
                    parent.id, childId);
        
        return new CodeBuilder()
            .append("%s.setAlignmentY(%ff);", childId, alignVert)
            .beginBlock("")
            .append("final JPanel $_border = new JPanel();")
            .append("$_border.setLayout(new OverlayLayout($_border));")
            .append("$_border.setBorder(BorderFactory.createEmptyBorder(" +
                        "%d, %<d, %<d, %<d));", gap)
            .append("$_border.setAlignmentY(%ff);", alignVert)
            .append("$_border.add(%s);", childId)
            .append("SPECIAL$%s$layoutChildren.add($_border);", parent.id)
            .append("SPECIAL$%s$layoutConstraints.add(null);", parent.id)
            .finishBlock()
            .finish();
    }
    
    private static final Set<String> imports = Builders.makeCollection(
            new TreeSet<String>(), 
            "javax.swing.BoxLayout",
            "java.awt.BorderLayout",
            "javax.swing.OverlayLayout",
            "javax.swing.BorderFactory",
            "javax.swing.JPanel"
    );
    
    public final Set<String> getImports() {
        return imports;
    }

    /* @see types.WidgetDef#getModifiers() */
    @Override
    public final List<String> getModifiers() {
        return Collections.emptyList();
    }

    @Override
    public final List<String> getVariables() {
        return Collections.singletonList("public BoxLayout " + id + ";\n");
    }
}