package lib.java;

import java.util.*;

import output.java.CodeBuilder;
import output.java.State;

import errors.BadAttributeException;
import errors.BadAttributeValueException;
import types.GUIAttributes;
import types.Widget;
import types.WidgetDef;
import util.Builders;

/*
 *  @EName Definition for a JFileChooser, both save and open.
 */
public final class FileChooser extends WidgetDef {

    private boolean needsFilterClass = false;
    
    public FileChooser(Widget widget, WidgetDef parent) {
        super(widget, parent);
    }

    @Override
    public String getLanguage() {
        return "Java";
    }

    @Override
    public String preamble() {
        return String.format("this.%s = new JFileChooser();\n", id);
    }

    private static final Map<String, String> synonyms =
        Builders.buildMap(new TreeMap<String, String>())
            .put("label", "title")
            .finish();
    
    @Override
    public Map<String, String> getSynonymousAttributes() {
        return synonyms;
    }

    private static final Map<String, List<String>> defaults =
        Builders.buildMultiMap(new TreeMap<String, List<String>>())
            .put("title", (String) null) 		//@EDefaultValue title null
            .put("directory", (String) null)	//@EDefaultValue directory null
            .put("mode", "either")				//@EDefaultValue mode either
            .put("acceptTypes")					//@EDefaultValue acceptTypes (empty)
            .put("allowAnyFile", "true")		//@EDefaultValue allowAnyFile true
            .put("allowMultiple", "false")		//@EDefaultValue allowMultiple false
            .put("showHidden", "false")         //@EDefaultValue showHidden false
            .finish();
    
    @Override
    public Map<String, List<String>> getDefaults() {
        return defaults;
    }

    private static enum Mode {
        files("FILES_ONLY"),
        directories("DIRECTORIES_ONLY"), 
        either("FILES_AND_DIRECTORIES");
        
        final String modeId;
        Mode(String modeId) {
            this.modeId = modeId;
        }
    }
    
    @Override
    public String addAttributes(GUIAttributes attrs)
            throws BadAttributeException {
        CodeBuilder builder = new CodeBuilder();
        
        //@EName title The title of the dialog box associated with the chooser.
        if (attrs.hasValue("title"))
            builder.append("%s.setDialogTitle(\"%s\");", 
                    id, attrs.getString("title"));
        
        //@EName directory The starting directory of the file chooser.
        if (attrs.hasValue("directory"))
            builder.append("%s.setCurrentDirectory(new File(\"%s\");",
                    id, attrs.getString("directory"));

        //@EName mode The mode of the File Chooser: either files, directories, or both.
        builder.append("%s.setFileSelectionMode(JFileChooser.%s);",
                id, attrs.getEnum("mode", Mode.class).modeId);
        
        //@EName acceptTypes The acceptable extensions for the File Chooser.
        final List<String> fileTypes = attrs.getStrings("acceptTypes");
        final Iterator<String> iter = fileTypes.iterator();
        
        while (iter.hasNext()) {
            // We're going to need the Filter inner class
            needsFilterClass = true;
            
            final String entry = iter.next();
            if (entry.startsWith("."))
                // It's an extension; use the default description
                builder.append("%s.addChoosableFileFilter(new Filter(\"%s\"));",
                        id, entry);
            else {
                // It's the description; it must be followed by an extension
                if (!iter.hasNext())
                    throw new BadAttributeValueException(
                            "file type description must be followed by the " +
                            "extension for the type", widget, "acceptTypes",
                            entry);
                final String ext = iter.next();
                if (!ext.startsWith("."))
                    throw new BadAttributeValueException(
                            "file extension must begin with a period",
                            widget, "acceptTypes", ext);
                
                builder.append("%s.addChoosableFileFilter(" +
                                "new Filter(\"%s\", \"%s\"));",
                        id, entry, ext);
            }
        }
        //@EName allowAnyFile Whether or not to allow all files.
        //
        //@EName allowMultiple Whether or not to allow multiple file selection.
        //
        //@EName showHidden Whether or not to show hidden files.
        //
        builder
            .append("%s.setAcceptAllFileFilterUsed(%s);",
                    id, attrs.getBoolean("allowAnyFile"))
            .append("%s.setMultiSelectionEnabled(%s);",
                    id, attrs.getBoolean("allowMultiple"))
            .append("%s.setFileHidingEnabled(%s);",
                    id, !attrs.getBoolean("showHidden"));
        
        return builder.toString();
    }
    
    private static final List<String> modifiers =
        Builders.makeCollection(new ArrayList<String>(),
            "public String getFileChooserTitle(JFileChooser chooser) {\n" +
            "    return chooser.getDialogTitle();\n" +
            "}\n",
                
            "public void setFileChooserTitle(" +
                    "JFileChooser chooser, String title) {\n" +
            "   chooser.setDialogTitle(title);\n" +
            "}\n",
            
            "public String getFileChooserDirectory(JFileChooser chooser) {\n" +
            "    return chooser.getCurrentDirectory().getPath();\n" +
            "}\n",
                
            "public void setFileChooserDirectory(" +
                    "JFileChooser chooser, String directory) {\n" +
            "   chooser.setCurrentDirectory(new File(directory));\n" +
            "}\n",
            
            "public FileChooserAnswer runFileChooserAsOpen(" +
                    "JFileChooser chooser, Component window) {\n" +
            "   switch (chooser.showOpenDialog(window)) {\n" +
            "       case JFileChooser.CANCEL_OPTION:\n" +
            "           return FileChooserAnswer.CANCEL;\n" +
            "       case JFileChooser.APPROVE_OPTION:\n" +
            "           return FileChooserAnswer.OK;\n" +
            "       default:\n" +
            "           return FileChooserAnswer.ERROR;\n" +
            "   }\n" +
            "}\n",
            
            "public FileChooserAnswer runFileChooserAsSave(" +
                    "JFileChooser chooser, Component window) {\n" +
            "   switch (chooser.showSaveDialog(window)) {\n" +
            "       case JFileChooser.CANCEL_OPTION:\n" +
            "           return FileChooserAnswer.CANCEL;\n" +
            "       case JFileChooser.APPROVE_OPTION:\n" +
            "           return FileChooserAnswer.OK;\n" +
            "       default:\n" +
            "           return FileChooserAnswer.ERROR;\n" +
            "   }\n" +
            "}\n",
            
            "public String getFileChooserSelectedFile(" +
                    "JFileChooser chooser) {\n" +
            "   return chooser.getSelectedFile().getPath();\n" +
            "}\n",
            
            "public String[] getFileChooserSelectedFiles(" +
                    "JFileChooser chooser) {\n" +
            "   final File[] files = chooser.getSelectedFiles();\n" +
            "   final String[] ans = new String[files.length];\n" +
            "   int ix = 0;\n" +
            "   for (File file : files)\n" +
            "       ans[ix++] = file.getPath();\n" +
            "   return ans;\n" +
            "}\n"
        );
    
    @Override
    public List<String> getModifiers() {
        return modifiers;
    }

    private static final Set<String> imports =
        Builders.makeCollection(new TreeSet<String>(),
                "javax.swing.JFileChooser",
                "java.io.File");
    
    private static final Set<String> filterClassImports =
        Builders.makeCollection(new TreeSet<String>(),
                "javax.swing.filechooser.*");
    
    @Override
    public Set<String> getImports() {
        if (!needsFilterClass)
            return imports;
        
        Set<String> ans = new TreeSet<String>(imports);
        ans.addAll(filterClassImports);
        return ans;
    }

    private static final String filterClassCode =
        new CodeBuilder(State.CLASS)
            .beginClass("protected final class Filter extends FileFilter")
            .append("protected final String ext;")
            .append("protected final String desc;")
            
            .beginMethod("protected Filter(String desc, String ext)")
            .append("this.desc = desc != null ? desc : ext + \" files\";")
            .append("this.ext = ext;")
            .finishMethod()
            
            .beginMethod("protected Filter(String ext)")
            .append("this(null, ext);")
            .finishMethod()
            
            .beginMethod("public String getDescription()")
            .append("return desc;")
            .finishMethod()
            
            .beginMethod("public boolean accept(File file)")
            .append("return file.getName().endsWith(ext);")
            .finishMethod()
            
            .finishClass()
            .finish();
    
    private static final String answerClassCode =
        new CodeBuilder(State.CLASS)
            .beginClass("public static enum FileChooserAnswer")
            .append("OK, CANCEL, ERROR")
            .finishClass()
            .finish();
    
    @Override
    public Set<String> getInnerClasses() {
        Set<String> ans = new TreeSet<String>();
        ans.add(answerClassCode);
        if (needsFilterClass)
            ans.add(filterClassCode);
        return ans;
    }
    
    @Override
    public List<String> getVariables() {
        return Collections.singletonList(
                String.format("public JFileChooser %s;\n", id));
    }

}
