package lib.java;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;
import java.util.TreeSet;

import errors.BadAttributeException;
import errors.BadChildWidgetException;
import types.GUIAttributes;
import types.Widget;
import types.WidgetDef;
import types.WidgetDef.Derived;
import util.Builders;

/**
 * @EName Frame Widget definition for a Frame element, like a JPanel in Java.
 * 
 * Attributes are:
 * - height
 * - width
 * - backgroundcolor
 * 
 * All other elements are acceptable children EXCEPT for:
 * - Window
 * - MenuBar
 * - Menu
 * - MenuItem
 * 
 * @author ritza
 *
 */
@Derived(Container.class)
public class Frame extends WidgetDef{

	public Frame(Widget widget, WidgetDef parent) {
		super(widget, parent);
	}

	@Override
	public String getLanguage() {
		return "Java";
	}

	@Override
	public String preamble() {
		StringBuilder builder = 
            new StringBuilder(id + " = new JPanel();\n");
        
        return builder.toString();
	}

    private static enum BorderType {
        none, etched;
    }
    
	/*
	 *  (non-Javadoc)
	 * @see types.WidgetDef#addAttributes(interpreter.GUIAttributes)
	 */
    	public String addAttributes(GUIAttributes attrs) 
            throws BadAttributeException {
		final StringBuilder builder = new StringBuilder(); //string to return
		// @EName height Defines the height of the frame
		// @EName width Defines the width of the frame
		if ( attrs.getInteger("width") != -1 &&
				attrs.getInteger("height") != -1)
			builder.append(String.format(
					"%s.setPreferredSize(new Dimension(%d, %d));\n", id,
					attrs.getInteger("width"),
					attrs.getInteger("height")));
		// @EName backgroundColor Defines the background color of the frame.
		if (attrs.getString("backgroundColor") != null)
			builder.append(String.format("%s.setBackground(%s);\n", id,
					attrs.getColor("backgroundColor")));
        /*
         * @EName borderType The type of border to draw. Currently "none" and
         * "etched" are supported values.
         */ 
        switch (attrs.getEnum("borderType", BorderType.class)) {
            case etched:
                builder.append(String.format(
                        "%s.setBorder(javax.swing.BorderFactory." +
                        "createEtchedBorder());\n", id));                    
        }
        
		return builder.toString();
	}

	/*
	 *  (non-Javadoc)
	 * @see types.WidgetDef#getDefaults()
	 * 
	 * This is where the defaults are defined. The idea is to have a
     * private static field that gets initialized at class loading time with
     * all the code; this way it doesn't get rebuilt each time. Then all that
     * getDefaults() has to do is return this field.
	 */
    private static final Map<String, List<String>> defaults =    
        Builders.buildMultiMap(new TreeMap<String, List<String>>())
			.put("height", "-1") 					//@EDefaultValue height (automatic)
			.put("width", "-1")						//@EDefaultValue width (automatic)
            .put("backgroundColor", (String)null)	//@EDefaultValue backgroundColor (system default)
            .put("borderType", "none")              //@EDefaultValue borderType none
            .finish();
    
    public Map<String,List<String>> getDefaults() {
		return defaults;
	}

	@Override
	public String addChild(String childType, 
                String childName, GUIAttributes childAttrs) 
            throws BadChildWidgetException, BadAttributeException {
		if (childType.equals("Window") || childType.equals("MenuBar")
				|| childType.equals("Menu") || childType.equals("MenuItem"))
			throw new BadChildWidgetException(widget);
		else if (isDerivedFrom(childType, Layout.class))
            return "";
		return String.format("%s.add(%s);\n", id, childName);
	}

	 /* This is where the modifier code is defined. The idea is to have a
     * private static field that gets initialized at class loading time with
     * all the code; this way it doesn't get rebuilt each time. Then all that
     * getModifiers() has to do is return this field. 
     * */
    
    private static final ArrayList<String> modifiers = Builders.makeCollection(
            new ArrayList<String>(),
            "public int getFrameHeight(JPanel frame) {\n" +
            "    return frame.getHeight();\n" +
            "}\n" +
            
            "public void setFrameHeight(JPanel frame, int newHeight) {\n" +
            "    Dimension newDimension = new Dimension(frame.getWidth(),newHeight);\n" +
            "    frame.setPreferredSize(newDimension);\n" +
            "}\n",
            
            "public int getFrameWidth(JPanel frame) {\n" +
            "    return frame.getWidth();\n" +
            "}\n" +
            
            "public void setFrameWidth(JPanel frame, int newWidth) {\n" +
            "    Dimension newDimension = new Dimension(newWidth,frame.getHeight());\n" +
            "    frame.setPreferredSize(newDimension);\n" +
            "}\n",
            
            "public Color getFrameBackgroundColor(JPanel frame) {\n" +
            "    return frame.getBackground();\n" +
            "}\n" +
            
            "public void setFrameBackgroundColor(JPanel frame, Color newBackgroundColor) {\n" +
            "    frame.setBackground(newBackgroundColor);\n" +
            "}\n"
    );
    
	public ArrayList<String> getModifiers() {
		return modifiers;
	}

	 /* This is where the import code is defined. The idea is to have a
     * private static field that gets initialized at class loading time with
     * all the code; this way it doesn't get rebuilt each time. Then all that
     * getImports() has to do is return this field. */
  
    private static final Set<String> imports = Builders.makeCollection(
            new TreeSet<String>(), 
            "import javax.swing.JPanel;\n",
            "import java.awt.Dimension;\n",
            "import java.awt.Color;\n"
    );

	public Set<String> getImports() {
		return imports;
	}

    @Override
    public List<String> getVariables() {
        return Collections.singletonList("public JPanel " + id + ";\n");
    }

}
