package lib.java;

import java.util.*;

import output.java.CodeBuilder;
import output.java.State;

import errors.BadAttributeException;
import errors.BadChildWidgetException;

import types.GUIAttributes;
import types.Widget;
import types.WidgetDef;
import types.WidgetDef.Derived;

import util.Builders;

/**
 * Widget Definition for a grid layout in Java.
 */

/* @EName GridLayout lays out its child widgets in a flexible grid, where each
 * widget occupies some cell in its upper-left corner and then spans some 
 * number of rows and columns. */

@Derived(Layout.class)
public class GridLayout extends WidgetDef {
    public GridLayout(Widget widget, WidgetDef parent) {
        super(widget, parent);
    }
    
    public String getLanguage() {
        return "Java";
    }
    
    public String preamble() {
        return String.format("%s = new GridBagLayout();\n", id);
    }
    
    private int rows = 0, cols = 0, gapHoriz = 0, gapVert = 0;
    private int currentRow = 0, currentCol = 0;
    private boolean[][] occupied = null;
    
    public String addAttributes(GUIAttributes attrs)
        throws BadAttributeException {
        rows = attrs.getInteger("rows");
        cols = attrs.getInteger("columns");
        gapHoriz = attrs.getInteger("gapHoriz");
        gapVert = attrs.getInteger("gapVert");
        if (rows != 0 && cols != 0)
            occupied = new boolean[rows][cols];
        return "";
    }
    
    private static final Map<String, List<String>> defaults = 
        Builders.buildMultiMap(new TreeMap<String, List<String>>())
                /*
                 * @EName rows The number of rows in the grid. By default, the
                 * grid has as many rows and columns as are needed to
                 * accomodate its child widgets. If explicit dimensions of the
                 * grid are given, the GridLayout will, by default,
                 * automatically place child widgets from left to right and top
                 * to bottom.
                 */
                .put("rows", "0")
                /*
                 * @EName columns The number of columns in the grid. See rows.
                 */
                .put("columns", "0")
                /*
                 * @EName gapHoriz The default horizontal gap between child
                 * widgets. Can be overridden for particular children by using
                 * the paddingLeft and paddingRight child attributes.
                 */
                .put("gapHoriz", "0")
                /*
                 * @EName gapVert The default vertical gap between child
                 * widgets. See gapHoriz.
                 */
                .put("gapVert", "0")
                .finish();
    
    public Map<String, List<String>> getDefaults() {
        return defaults;
    }
    
    private static final Map<String, List<String>> childDefaults =
        Builders.buildMultiMap(new TreeMap<String, List<String>>())
                /*
                 * @EName row The row in which to place the child. By default,
                 * if the columns attribute is set for the GridLayout and the
                 * last widget was placed in the rightmost column, this one
                 * will be placed in the next row; if the last widget was not
                 * in the rightmost column, or the columns attribute is not
                 * given, this widget will be placed in the same row as the
                 * last.
                 */
                .put("row", (String) null)
                /*
                 * @EName column The column in which to place the child. By
                 * default, if the columns attribute is set for the GridLayout
                 * and the last widget was placed in the rightmost column, this
                 * one will be placed in the next row in the first column (that
                 * is, the widget will "wrap" to the next row); if the last
                 * widget was not in the rightmost column, or the columns
                 * attribute is not given, this widget will be placed
                 * immediately to the right of the last.
                 */
                .put("column", (String) null)
                /*
                 * @EName spanRows The number of rows spanned by the widget, or
                 * alternatively the height of the widget in rows. Note that if
                 * spanRows is set for a widget, it may overlap with a widget
                 * placed below it if explicit row and column attributes are
                 * not given. (The automatic placement algorithm does not take
                 * span into account.) 
                 * <p>
                 * By default, each widget takes up a single cell; that is,
                 * spanRows == spanColumns == 1.
                 */
                .put("spanRows", "1")
                /*
                 * @EName spanColumns The number of columns spanned by the
                 * widget, or alternatively the width of the widget in rows.
                 * Note that if spanColumns is set for a widget, it may overlap
                 * with the next widget placed if row and column attributes are
                 * not given. (The automatic placement algorithm does not take
                 * span into account.)
                 * <p>
                 * By default, each widget takes up a single cell; that is,
                 * spanRows == spanColumns == 1.
                 */
                .put("spanColumns", "1")
                /*
                 * @EName weightHoriz The horizontal weight of the widget.
                 * Should be a real number between 0.0 and 1.0. If there is
                 * extra space in the grid after all widgets are placed, the
                 * remaining space is allocated in proportion to the weights of
                 * the widgets. The default is 1.0, but see fillHoriz.
                 */
                .put("weightHoriz", "1.0")
                /*
                 * @EName weightVert The vertical weight of the widget. Should
                 * be a real number between 0.0 and 1.0. See weightHoriz.
                 */
                .put("weightVert", "1.0")
                /*
                 * @EName alignHoriz The horizontal alignment of the widget.
                 * Should be "left", "center", or "right". The default is
                 * "center".
                 */
                .put("alignHoriz", "center")
                /*
                 * @EName alignVert The vertical alignment of the widget.
                 * Should be "top", "center", or "bottom". The default is
                 * "center".
                 */
                .put("alignVert", "center")
                /*
                 * @EName fillHoriz If true, the widget will expand
                 * horizontally to fit any extra room allocated to it. The
                 * default is false, so the widget will tend to keep its
                 * natural size.
                 */
                .put("fillHoriz", "true")
                /*
                 * @EName fillVert If true, the widget will expand vertically
                 * to fit any extra room allocated to it. The default is false,
                 * so the widget will tend to keep its natural size.
                 */
                .put("fillVert", "true")
                /*
                 * @EName marginLeft The amount of space between the left side
                 * of the widget and the bounding box of its cell in the grid.
                 */
                .put("marginLeft", "0")
                /*
                 * @EName marginRight The amount of space between the right
                 * side of the widget and the bounding box of its cell in the
                 * grid.
                 */
                .put("marginRight", "0")
                /*
                 * @EName marginTop The amount of space between the top of the
                 * widget and the bounding box of its cell in the grid.
                 */
                .put("marginTop", "0")
                /*
                 * @EName marginBottom The amount of space between the bottom
                 * of the widget and the bounding box of its cell in the grid.
                 */
                .put("marginBottom", "0")
                /*
                 * @EName paddingWidth The amount of extra space to put in each
                 * of the left and right gutters next to the widget's cell in
                 * the grid, between the cell and the neighboring cells.
                 */
                .put("paddingWidth", "0")
                /*
                 * @EName paddingHeight The amount of extra space to put in
                 * each of the top and bottom gutters adjacent to the widget's
                 * cell in the grid, between the cell and the neighboring
                 * cells.
                 */
                .put("paddingHeight", "0")
                .finish();
    
    @Override
    public Map<String, List<String>> getChildDefaults() {
        return childDefaults;
    }
    
    private enum HorizAlignment {
        left, center, right
    }
    
    private enum VertAlignment {
        top, center, bottom
    }
    
    @Override
    public String addChild(String childType, String childId,
            GUIAttributes attrs) throws BadChildWidgetException,
            BadAttributeException {
        CodeBuilder builder = new CodeBuilder(State.METHOD);
        
        // Start a new scope to avoid id collisions with other local vars
        builder.beginBlock("");
        
        String constraintsId = String.format("$loc_%s_%s_constraints",
                id, childId);
        builder.append(
                "GridBagConstraints %s = new GridBagConstraints();\n",
                constraintsId);
        
        // row and column
        
        {
            final int row;
            final int column;
            
            if (attrs.hasValue("row")) {
                row = attrs.getInteger("row") - 1;
                if (row != currentRow) {
                    currentRow = row;
                    if (attrs.hasValue("column"))
                        column = (currentCol = attrs.getInteger("column")) - 1;
                    else
                        column = (currentCol = 1) - 1;
                } else
                    if (attrs.hasValue("column"))
                        column = (currentCol = attrs.getInteger("column")) - 1;
                    else
                        // TODO Implement the fancy-pants occupied[][] business
                        // I started ...
                        column = currentCol++;
            } else {
                if (attrs.hasValue("column")) {
                    row = currentRow;
                    column = (currentCol = attrs.getInteger("column")) - 1;
                } else if (cols == 0 || currentCol < cols) {
                    row = currentRow;
                    column = currentCol++;
                } else {
                    row = ++currentRow;
                    if (row >= rows) {
                        // Do nothing; maybe someday we'll be able to raise a
                        // warning of some sort
                    }                        
                    column = (currentCol = 1) - 1;
                }
            }
            
            if (occupied != null)
                try {
                    occupied[row][column] = true;
                } catch (ArrayIndexOutOfBoundsException e) {
                    // Do nothing. If they explicitly set a location that's out
                    // of bounds, we're not going to mess with that.
                }
            
            if (row != -1)
                builder.append("%s.gridy = %d;\n", constraintsId, row);
            if (column != -1)
                builder.append("%s.gridx = %d;\n", constraintsId, column);
        }
        
        // spanRows and spanColumns
        
        {
            final int rows =
                attrs.getString("spanRows").equalsIgnoreCase("ALL") ? 
                        0 : attrs.getInteger("spanRows");
            final int cols =
                attrs.getString("spanColumns").equalsIgnoreCase("ALL") ? 
                        0 : attrs.getInteger("spanColumns");
            builder.append(
                    "%s.gridheight = %d;\n" + "%1$s.gridwidth = %d;\n",
                    constraintsId, rows, cols);
        }
            
        // weight
        
        builder.append(
                "%s.weightx = %f;\n" + "%1$s.weighty = %f;\n",
                constraintsId,
                attrs.getDouble("weightHoriz"),
                attrs.getDouble("weightVert")
                );
        
        // align
        
        {
            final String anchorId;
            OUTER:
            switch(attrs.getEnum("alignHoriz", HorizAlignment.class)) {
            case left:
                switch (attrs.getEnum("alignVert", VertAlignment.class)) {
                case top: anchorId = "FIRST_LINE_START"; break OUTER;
                case center: anchorId = "LINE_START"; break OUTER;
                case bottom: anchorId = "LAST_LINE_START"; break OUTER;
                }
            case center:
                switch (attrs.getEnum("alignVert", VertAlignment.class)) {
                case top: anchorId = "PAGE_START"; break OUTER;
                case center: anchorId = "CENTER"; break OUTER;
                case bottom: anchorId = "PAGE_END"; break OUTER;
                }
            case right:
                switch (attrs.getEnum("alignVert", VertAlignment.class)) {
                case top: anchorId = "FIRST_LINE_END"; break OUTER;
                case center: anchorId = "LINE_END"; break OUTER;
                case bottom: anchorId = "LAST_LINE_END"; break OUTER;
                }
            default:
                // Should never be reached
                anchorId = "CENTER /* default by error in GridLayout */";
            }
        
            builder.append(
                    "%s.anchor = GridBagConstraints.%s;\n",
                    constraintsId, anchorId);
        }
        
        // fill
        
        {
            final String fillId;
            if (attrs.getBoolean("fillHoriz"))
                if (attrs.getBoolean("fillVert"))
                    fillId = "BOTH";
                else
                    fillId = "HORIZONTAL";
            else
                if (attrs.getBoolean("fillVert"))
                    fillId = "VERTICAL";
                else
                    fillId = "NONE";
            
            builder.append(
                    "%s.fill = GridBagConstraints.%s;\n",
                    constraintsId, fillId);
        }
        
        // margin
        
        final int extraHoriz = gapHoriz / 2;
        final int extraVert = gapVert / 2;
        
        builder.append(
                "%s.insets = new Insets(%d, %d, %d, %d);\n",
                constraintsId,
                attrs.getInteger("marginTop") + extraVert,
                attrs.getInteger("marginLeft") + extraHoriz,
                attrs.getInteger("marginBottom") + extraVert,
                attrs.getInteger("marginRight") + extraHoriz)
            .append( // padding
                "%s.ipadx = %d;\n" + "%1$s.ipady = %d;\n",
                constraintsId,
                attrs.getInteger("paddingWidth"),
                attrs.getInteger("paddingHeight"))

            .append(
                "SPECIAL$%s$layoutConstraints.add(%s);\n",
                parent.id, constraintsId);
        
        builder.finishBlock();
        
        builder.append("SPECIAL$%s$layoutChildren.add(%s);\n",
                parent.id, childId);
        
        return builder.toString();
    }
    
    private static final Set<String> imports = Builders.makeCollection(
            new TreeSet<String>(), 
            "import java.awt.GridBagLayout;\n",
            "import java.awt.GridBagConstraints;\n",
            "import java.awt.Insets;\n"
    );
    
    public Set<String> getImports() {
        return imports;
    }

    /* @see types.WidgetDef#getModifiers() */
    @Override
    public List<String> getModifiers() {
        return Collections.emptyList();
    }

    @Override
    public List<String> getVariables() {
        return Collections.singletonList("public GridBagLayout " + id + ";\n");
    }
}
