package lib.java;

import java.util.*;

import errors.BadAttributeException;
import errors.BadChildWidgetException;
import types.GUIAttributes;
import types.Widget;
import types.WidgetDef;
import types.WidgetDef.Derived;
import util.Builders;

/**
 * Menu definition for a Java JMenu.  A Menu is contained by a MenuBar, 
 * and contains MenuItems.  A Menu's attributes include:
 *  - label
 *  - mnemonic
 *  
 * @EName Definition for a Java JMenu.  A Menu is contained by a MenuBar, 
 * and contains MenuItems.  
 *
 */
// @EName Menu Makes a menu to be contained by a MenuBar. Items of a
//        Menu are MenuItems.
@Derived(Clickable.class)
public class Menu extends WidgetDef {

	public Menu(Widget widget, WidgetDef parent) {
        super(widget, parent);
    }
    
	@Override
	public String getLanguage() {
		return "Java";
	}

	@Override
	public String preamble() {
		return id + " = new JMenu();\n";
	}

	@Override
	public String postamble() {
		return "";
	}

	@Override
	public String addAttributes(GUIAttributes attrs) throws BadAttributeException {
		final StringBuilder builder = new StringBuilder();

        // @EName label Sets the menu's title.
        if (attrs.hasValue("label"))
            builder.append(String.format("%s.setText(\"%s\");\n",id,
                    attrs.getString("label")));
		// @EName mnemonic When set to default, is not included in attributes.
		if (attrs.hasValue("mnemonic"))
			builder.append(String.format("%s.setMnemonic(%s);\n",id,
					"KeyEvent.VK_".concat(attrs.getOne("mnemonic"))));
		return builder.toString();
	}

	/*
	 * This is where the defaults are defined. The idea is to have a
     * private static field that gets initialized at class loading time with
     * all the code; this way it doesn't get rebuilt each time. Then all that
     * getDefaults() has to do is return this field. 
	 */
	
	private static final Map<String, List<String>> defaults =    
        Builders.buildMultiMap(new TreeMap<String, List<String>>())
      .put("label", (String) null)     // @EDefaultValue label null
      .put("mnemonic", (String) null)  // @EDefaultValue mnemonic null
      .finish();
	
	@Override
	public Map<String, List<String>> getDefaults() {
		return defaults; 
	}

    private static final Map<String, String> synonyms =
        Builders.buildMap(new TreeMap<String, String>())
            .put("text", "label")
            .finish();
    
    @Override
    public Map<String, String> getSynonymousAttributes() {
        return synonyms;
    }
    
	@Override
	public String addChild(String childType, String childName, 
                GUIAttributes childAttrs) 
            throws BadChildWidgetException {
		childType = childType.toLowerCase();
		if (childType.equals("menuitem") || childType.equals("menuseparator"))
			return id.concat(".add(").concat(childName).concat(");\n");
		throw new BadChildWidgetException(widget);
	}
	
	 /* This is where the modifier code is defined. The idea is to have a
     * private static field that gets initialized at class loading time with
     * all the code; this way it doesn't get rebuilt each time. Then all that
     * getModifiers() has to do is return this field. */
	
	private static final ArrayList<String> modifiers = Builders.makeCollection(
			new ArrayList<String>(),
			"public String getMenuText(JMenu menu) {\n" +
			"\treturn menu.getText();\n" +
			"}\n" +
			
			"public void setMenuText(JMenu menu, String newText) {\n" +
			"\tmenu.setText(newText);\n" +
			"}\n",
			
			"public int getMenuMnemonic(JMenu menu) {\n" +
			"\treturn menu.getMnemonic();\n" +
			"}\n" +
			
			"public void setMenuMnemonic(JMenu menu, int newMnemonic) {\n" +
			"\tmenu.setMnemonic(newMnemonic);\n" +
			"}\n"
		);

	@Override
	public ArrayList<String> getModifiers() {
		return modifiers;
	}
	
	/* This is where the import code is defined. The idea is to have a
     * private static field that gets initialized at class loading time with
     * all the code; this way it doesn't get rebuilt each time. Then all that
     * getImports() has to do is return this field. */
	 private static final Set<String> imports = Builders.makeCollection(
	            new TreeSet<String>(), 
	            "import javax.swing.JMenu;\n",
	            "import java.awt.event.KeyEvent;\n");

	@Override
	public Set<String> getImports() {
		return imports;
	}

    @Override
    public List<String> getVariables() {
        return Collections.singletonList("public JMenu " + id + ";\n");
    }
}