package lib.java;

import java.util.*;

import errors.BadAttributeException;
import errors.BadChildWidgetException;
import types.GUIAttributes;
import types.Widget;
import types.WidgetDef;
import util.Builders;

/**
 * @EName Scrolling Frame widget from Java's JScrollPane.
 * 
 * If a GridLayout is being used, then there needs to be
 * a way to set the ScrollingFrame size.  Therefore, the only 
 * two attributes are:
 * - height
 * - width
 * 
 * The ScrollingFrame accepts all children EXCEPT:
 * - Window
 * - MenuBar
 * - Menu
 * - MenuItem
 * - GridLayout
 * 
 * @author ritza
 *
 */
public class ScrollingFrame extends WidgetDef{
	private boolean oneChild = false;
	
	public ScrollingFrame(Widget widget, WidgetDef parent) {
        super(widget, parent);
    }

	@Override
	public String getLanguage() {
		return "Java";
	}

	@Override
	public String preamble() {
		return id + " = new JScrollPane();\n";
	}

	@Override
	public String postamble() {
		return "";
	}

	@Override
	public String addAttributes(GUIAttributes attributes) throws BadAttributeException {
		final StringBuilder builder = new StringBuilder(); //string to return
		// @EName height Height for the ScrollingFrame.
		//@EName width Width for the ScrollingFrame.
		if ( attributes.getInteger("width") != -1 &&
				attributes.getInteger("height") != -1)
			builder.append(String.format(
					"%s.setPreferredSize(new Dimension(%d, %d));\n", id,
					attributes.getInteger("width"),
					attributes.getInteger("height")));
		return builder.toString();
	}

	/*
	 *  (non-Javadoc)
	 * @see types.WidgetDef#getDefaults()
	 * 
	 * This is where the defaults are defined. The idea is to have a
     * private static field that gets initialized at class loading time with
     * all the code; this way it doesn't get rebuilt each time. Then all that
     * getDefaults() has to do is return this field. 
     * 
     * TODO: what should we set height and width defaults to?
	 */
    
    private static final Map<String, List<String>> defaults =    
        Builders.buildMultiMap(new TreeMap<String, List<String>>())
			.put("height", "-1") 	//@EDefaultValue height -1
			.put("width", "-1")		//@EDefaultValue width -1
            .finish();
	@Override
	public Map<String, List<String>> getDefaults() {
		return defaults;
	}

	@Override
	//TODO:  How do we ensure that ScrollingFrame only has 1 child??  I currently
	// have a boolean.  It also throws an inappropriate exception.
	public String addChild(String childType, String childName, 
                GUIAttributes childAttrs)
            throws BadChildWidgetException, BadAttributeException {
		if (childType.equals("Window") || childType.equals("MenuBar")
				|| childType.equals("Menu") || childType.equals("MenuItem")
				|| isDerivedFrom(childType, Layout.class) || oneChild)
			throw new BadChildWidgetException(widget);
		oneChild = true;
		return String.format("%s.setViewportView(%s);\n", id, childName);
	}
	
	 /* This is where the modifier code is defined. The idea is to have a
     * private static field that gets initialized at class loading time with
     * all the code; this way it doesn't get rebuilt each time. Then all that
     * getModifiers() has to do is return this field. 
     * */
	
	private static final ArrayList<String> modifiers = Builders.makeCollection(
            new ArrayList<String>(),
            "public int getScrollingFrameHeight(JScrollPane scrollingFrame) {\n" +
            "    return scrollingFrame.getHeight();\n" +
            "}\n" +
            
            "public void setScrollingFrameHeight(JScrollPane scrollingFrame, int newHeight) {\n" +
            "    Dimension newDimension = new Dimension(scrollingFrame.getWidth(),newHeight);\n" +
            "    scrollingFrame.setPreferredSize(newDimension);\n" +
            "}\n",
            
            "public int getScrollingFrameWidth(JScrollPane scrollingFrame) {\n" +
            "    return scrollingFrame.getWidth();\n" +
            "}\n" +
            
            "public void setScrollingFrameWidth(JScrollPane scrollingFrame, int newWidth) {\n" +
            "    Dimension newDimension = new Dimension(newWidth,scrollingFrame.getHeight());\n" +
            "    scrollingFrame.setPreferredSize(newDimension);\n" +
            "}\n");

	@Override
	public List<String> getModifiers() {
		return modifiers;
	}

	 /* This is where the import code is defined. The idea is to have a
     * private static field that gets initialized at class loading time with
     * all the code; this way it doesn't get rebuilt each time. Then all that
     * getImports() has to do is return this field. */
  
    private static final Set<String> imports = Builders.makeCollection(
            new TreeSet<String>(), 
            "import javax.swing.JScrollPane;\n",
            "import java.awt.Dimension;\n"
    );

	public Set<String> getImports() {
		return imports;
	}

    @Override
    public List<String> getVariables() {
        return Collections.singletonList("public JScrollPane " + id + ";\n");
    }

}
