package lib.java;


import java.util.*;

import javax.swing.*;

import output.java.CodeBuilder;

import errors.BadAttributeException;
import errors.BadAttributeValueException;
import errors.BadChildWidgetException;

import types.GUIAttributes;
import types.Widget;
import types.WidgetDef;
import types.WidgetDef.*;

import util.Builders;

/**
 * @EName Definition for a Java JFrame.
 */

@Derived(Container.class)
public class Window extends WidgetDef {
	
	public Window(Widget widget, WidgetDef parent) {
        super(widget, parent);
    }
	
	/*
	 *  (non-Javadoc)
	 * @see types.WidgetDef#getLanguage()
	 */
	public String getLanguage() {
		return "Java";
	}
	
	/*
	 *  (non-Javadoc)
	 * @see types.WidgetDef#preamble()
	 */
	public String preamble() {
        return String.format("%s = new JFrame();\n", id);
	}
	
	/*
	 *  (non-Javadoc)
	 * @see types.WidgetDef#postamble()
	 */
	public String postamble() {
        CodeBuilder builder = new CodeBuilder();
        builder.beginBlock("if (!AUX$%s$hasLayout)", id)
                .append("%s.setLayout(new FlowLayout());\n", id)
                .finishBlock()
        
                .append("%1$s.setVisible(true);\n" + 
                        "%1$s.pack();\n", id);

		return builder.finish();
	}
	
	/*
	 *  (non-Javadoc)
	 * @see types.WidgetDef#addAttributes(interpreter.GUIAttributes)
	 */
	public String addAttributes(GUIAttributes attrs) throws 
	BadAttributeException {
		final StringBuilder builder = new StringBuilder(); //string to return
		//@EName title Sets the Window's title.
        if (attrs.getOne("title") != null)
            builder.append(String.format("%s.setTitle(\"%s\");\n", id, 
                    attrs.getOne("title")));
		//@EName height Sets the Window's height.
        //@EName width Sets the Window's width.
        if (attrs.getInteger("width") != -1 &&
                attrs.getInteger("height") != -1)
            builder.append(String.format(
                    "%s.setPreferredSize(new Dimension(%d, %d));\n", id,
                    attrs.getInteger("width"),
                    attrs.getInteger("height")));
		//@EName resizable Boolean to set the Window to be resizable (true) or not resiable (false).
		builder.append(String.format("%s.setResizable(%s);\n",
		        id, attrs.getBoolean("resizable")));
		//@EName backgroundColor sets the BackgroundColor for the Window.
        if (attrs.getOne("backgroundColor") != null)
            builder.append(String.format("%s.setBackground(%s);\n", id,
                    attrs.getColor("backgroundColor")));
		//@EName onClose Sets the action when the Window is closed.  Options
        // are donothing, hide, dispose, and exit.
		{
		    final String close =
		        attrs.getOne("onClose").toLowerCase();
		    builder.append(id + ".setDefaultCloseOperation(");
		    if (close.equals("donothing") || close.equals("do nothing"))
		        builder.append(WindowConstants.DO_NOTHING_ON_CLOSE);
		    else if (close.equals("hide"))
		        builder.append(WindowConstants.HIDE_ON_CLOSE);
		    else if (close.equals("dispose"))
		        builder.append(WindowConstants.DISPOSE_ON_CLOSE);
		    else if (close.equals("exit"))
		        builder.append(WindowConstants.EXIT_ON_CLOSE);
		    else throw new BadAttributeValueException(
		            "Expected \"do nothing\", \"hide\", \"dispose\", or " +
		            "\"exit\"", widget, "onClose", close);
		    builder.append(");\n");
		}
		return builder.toString();
	}
	
	/*
	 * @see types.WidgetDef#getDefaults()
	 * 
	 * This is where the defaults are defined. The idea is to have a
     * private static field that gets initialized at class loading time with
     * all the code; this way it doesn't get rebuilt each time. Then all that
     * getDefaults() has to do is return this field. 
	 */
    private static final Map<String, List<String>> defaults =    
        Builders.buildMultiMap(new TreeMap<String, List<String>>())
            .put("title", (String) null) 	//@EDefaultValue title null
			.put("height", "-1")			//@EDefaultValue height -1
			.put("width", "-1")				//@EDefaultValue width -1
            .put("resizable", "true")		//@EDefaultValue resizable true
            .put("backgroundColor", (String) null)	//@EDefaultValue backgroundColor null
            .put("onClose", "exit")			//@EDefaultValue onClose exit
            .finish();
    
    public Map<String,List<String>> getDefaults() {
		return defaults;
	}
	
    private static final Map<String, String> synonyms =
        Builders.buildMap(new TreeMap<String, String>())
            .put("label", "title")
            .finish();
    
    @Override
    public Map<String, String> getSynonymousAttributes() {
        return synonyms;
    }
    
    @Override
	public String addChild(String childType, String childId, 
                GUIAttributes childAttrs)
            throws BadChildWidgetException {
        if (isDerivedFrom(childType, Layout.class))
            return "";
        if (isDerivedFrom(childType, MenuBar.class))
            return String.format("%s.setJMenuBar(%s);\n", id, childId);
        return String.format("%s.add(%s);\n", id, childId);
    }
    
    /*
     * This is where the modifier code is defined. The idea is to have a
     * private static field that gets initialized at class loading time with
     * all the code; this way it doesn't get rebuilt each time. Then all that
     * getModifiers() has to do is return this field.
     */
    
    private static final ArrayList<String> modifiers = Builders.makeCollection(
            new ArrayList<String>(),
            "public String getWindowTitle(JFrame window) {\n" +
            "    return window.getTitle();\n" +
            "}\n"+
            
            "public void setWindowTitle(JFrame window, String newTitle) {\n" +
            "    window.setTitle(newTitle);\n" +
            "}\n",
            
            "public int getWindowHeight(JFrame window) {\n" +
            "    return window.getHeight();\n" +
            "}\n" +
            
            "public void setWindowHeight(JFrame window, int newHeight) {\n" +
            "    Dimension newDimension = new Dimension(window.getWidth(),newHeight);\n" +
            "    window.setPreferredSize(newDimension);\n" +
            "}\n",
            
            "public int getWindowWidth(JFrame window) {\n" +
            "    return window.getWidth();\n" +
            "}\n" +
            
            "public void setWindowWidth(JFrame window, int newWidth) {\n" +
            "    Dimension newDimension = new Dimension(newWidth,window.getHeight());\n" +
            "    window.setPreferredSize(newDimension);\n" +
            "}\n",
            
            "public boolean getWindowResizable(JFrame window) {\n" +
            "    return window.isResizable();\n" +
            "}\n" +
            
            "public void setWindowResizable(JFrame window, boolean newResizable) {\n" +
            "    window.setResizable(newResizable);\n" +
            "}\n",
            
            "public Color getWindowBackgroundColor(JFrame window) {\n" +
            "    return window.getBackground();\n" +
            "}\n" +
            
            "public void setWindowBackgroundColor(JFrame window, Color newBackgroundColor) {\n" +
            "    window.setBackground(newBackgroundColor);\n" +
            "}\n",
            
            "public int getWindowOnClose(JFrame window) {\n" +
            "    return window.getDefaultCloseOperation();\n" +
            "}\n" +
            
            "public void setWindowOnClose(JFrame window, int newClose) {\n" +
            "    window.setDefaultCloseOperation(newClose);\n" +
            "}\n",
            
            "public LayoutManager getWindowLayout(JFrame window) {\n" +
            "    return window.getLayout();\n" +
            "}\n" +
            
            "public void setWindowLayout(JFrame window, LayoutManager newLayout) {\n" +
            "    window.setLayout(newLayout);\n" +
            "}\n"
    );
    
	public ArrayList<String> getModifiers() {
		return modifiers;
	}

	 /* This is where the import code is defined. The idea is to have a
     * private static field that gets initialized at class loading time with
     * all the code; this way it doesn't get rebuilt each time. Then all that
     * getImports() has to do is return this field. */
  
    private static final Set<String> imports = Builders.makeCollection(
            new TreeSet<String>(), 
            // Note that the word "import" and the semicolon and newline are
            // no longer required (though they're accepted for backward
            // compatibility).
            "java.util.*",
            "java.awt.Component",
            "javax.swing.JFrame",
            "java.awt.Dimension",
            "java.awt.Color",
            "java.awt.FlowLayout",
            "java.awt.LayoutManager"
    );

	public Set<String> getImports() {
		return imports;
	}

    @Override
    public List<String> getVariables() {
        return Collections.singletonList(
                String.format("public JFrame %s;\n", id));
    }
}
