package lib.lang;

import java.util.regex.*;

import output.java.CodeBuilder;

import errors.BadWidgetTypeException;
import errors.GUIFileException;

import types.ErrorHandler;
import types.EventHandlerDef;
import types.GUIFile;
import types.LangDef;
import types.Widget;
import types.WidgetDef;
import types.GUIFile.EventHandler;
import types.GUIFile.Flag;

/**
 * Language definition for compiling Elvis code into Java source.
 */
public class Java extends LangDef {
	
	/**
	 * Default constructor. Calls the superclass constructor with "lib.java",
     * the prefix for Java widget definitions, as the argument.
	 */
	public Java() {
        super("lib.java");
	}
    
	/** generateCode implements the LangDef specification, accepting a guiFile
	 * and an error handler and returning a double array of code file names and 
	 * content to be outputted in whatever way desired by the Assembler.
	 * 
	 * @param guiFile the electronic representation of the code to be generated
	 * @param errorHandler the errorHandler to be consulted in the event of a
	 * 		guiFile exception.
	 * @return files a double array of file names and contents for the code
	 */
    public String[][] generateCode(GUIFile guiFile, ErrorHandler errorHandler)
            throws GUIFileException {
        
        if (errorHandler != null)
            this.errorHandler = errorHandler;
        
		String projectName = guiFile.getProjectName();
        final String controllerId = guiFile.getControllerId();
        
		/* access guiFile for these booleans */
		boolean generateController = !guiFile.handlers.isEmpty(); 

        final CodeBuilder builder = new CodeBuilder();
        
		appendPreamble(builder, guiFile,
                guiFile.flags.contains(Flag.GENERATE_MAIN),
                generateController);
		
		for (Widget widget : guiFile.widgets)
            try {
                builder.append(createWidget(widget));
            } catch (GUIFileException e) {
                if (!errorHandler.handleError(e))
                    throw e;
            }
			
		/*
         * This closes off the main method or constructor of the generated
         * code, depending on whether it was executable or instantiable.
         */
        builder.finishMethod();
		
        appendVariables(builder);
        
		if (!guiFile.flags.contains(Flag.SUPPRESS_MODIFIERS))
			appendModifiers(builder);
		
		appendInnerClasses(builder);
	
		builder.finishClass();

		addImports(builder);
		
		String[][] files;
		if(!generateController){
			files = new String[1][2];
			files[0][0] = projectName + ".java";
			files[0][1] = builder.toString();
		} else {
			files = new String[2][2];
			files[0][0] = projectName + ".java";
			files[0][1] = builder.toString();
			files[1][0] = controllerId + ".java";
			files[1][1] = getController(guiFile);
		}
		
		return files;
	}
    /** returns the root widget type of Java's widget library.
     * @return lib.java.Widget.class the class of the root widget.
     */
    @Override
    public Class<? extends WidgetDef> getRootWidgetType() {
        return lib.java.Widget.class;
    }


    /*
     * NOTE: This function begins a constructor. Therefore it leaves
     * an open brace hanging, which MUST be closed.
	 */ 
    /** appendPreamble creates the main method and the constructor of the project's
     * main class.
     * @param builder the CodeBuilder object associated with the file
     * @param guiFile the GUIFile object passed in by the Assembler
     * @param generateMain a boolean specifying whether a main method is desired
     * @param generateController a boolean specifying whether a controller is specified
     */
    private void appendPreamble(CodeBuilder builder,
            GUIFile guiFile, boolean generateMain, boolean generateController) 
			throws BadWidgetTypeException {
        final String projectName = guiFile.getProjectName();
        final String controllerId = guiFile.getControllerId();
        final String packageId = guiFile.getPackageId();
        
        if (packageId != null && !packageId.equals(""))
            builder.addPackageStatement(packageId);

        builder.beginClass("public class " + projectName);
			
		if (generateMain) {
            builder.beginMethod("public static void main(String[] args)");

            if(generateController)
            	    builder.append("new %s(null);", projectName);
            else
                builder.append("new %s();", projectName);
            
            builder.finishMethod();
        }
		
        if (generateController)
            builder.append("public final %s controller;", controllerId)
        		    .beginMethod("public %s(%s c)", projectName, controllerId)
        		    .append("controller = c;");
        else
            builder.beginMethod("public %s()", projectName);
  	}
    
    /** adds all of the imports gathered after all the widgets are processed.
     * @param builder the CodeBuilder object associated with the file.
     */
	private void addImports(CodeBuilder builder) {
		for (String importStmt : imports) {
            // Stopgap measure until things are converted to leave off the
            // "import " part and the newline
			importStmt = importStmt
                .replaceAll("^import ", "")
                .replaceAll(";\n?$", "");
            builder.addImport(importStmt);
        }	
	}
	
	/** appends all of the modifiers gathered as the widgets were processed.
	 * @param builder the CodeBuilder object associated with the file.
	 */
	private void appendModifiers(CodeBuilder builder) {
		for (String modifier : modifiers)
			builder.append(modifier);
	}
	
	/** appends all of the inner classes gathered as the widgets were processed.
	 * @param builder the CodeBuilder object associated with the file.
	 */
	private void appendInnerClasses(CodeBuilder builder) {
		for (String innerClass : innerClasses)
			builder.append(innerClass);
	}
	
    private static final Pattern oneLineComment = Pattern.compile("^//");
    private static final Pattern multiCommentStart = Pattern.compile("^/\\*");
    private static final Pattern multiCommentEnd = Pattern.compile("\\*/$");
    private static final Pattern anonPattern = Pattern.compile("\\banon\\$");
    private static final Pattern visPattern = Pattern.compile(
            "\\bpublic\\b|\\bprivate\\b|\\bprotected\\b");
    /**
     * 	Checks whether or not a variable is declared with an anonymous name
     * and changes it's visibility to private.
	 * @param builder the CodeBuilder object associated with the file.
	 */
	private void appendVariables(CodeBuilder builder) {
        boolean inComment = false;
		for (String variable : variables) {
            if (inComment && multiCommentEnd.matcher(variable).find())
                inComment = false;
            else if (!inComment && 
                    multiCommentStart.matcher(variable).find())
                inComment = true;
            
            if (!inComment && !oneLineComment.matcher(variable).find() &&
                    anonPattern.matcher(variable).find()) {
                Matcher visMatcher = visPattern.matcher(variable); 
                if (visMatcher.find()) {
                    variable = visMatcher.replaceFirst("private");
                } else {
                    variable = "private " + variable;
                }
            }
			builder.append(variable);
        }
	}
	/**
	 * generates the controller file for a Java controller interface.
	 * @param guiFile the GUIFile object associated with the file.
	 * @return builder.toString() the string describing the Controller interface
	 */
	private String getController(GUIFile guiFile) {
        final String controllerId = guiFile.getControllerId();
        final CodeBuilder builder = new CodeBuilder();
		final String packageId = guiFile.getPackageId();
        
        if (packageId != null)
            builder.addPackageStatement(packageId);
        
        builder.beginClass("public interface %s", controllerId);
        
		for (EventHandler handler : guiFile.handlers) {
            EventHandlerDef handlerDef = EventHandlerDef.valueOf(
                    handler.handlerType);
				builder.format("public void %s(%s);",
                        handler.id, handlerDef.handlerSig);
                for (String imp : handlerDef.imports)
                    builder.addImport(imp);
        }
		
        builder.finishClass();
       
        return builder.toString();
	}
	
}
