package output.java;

import java.util.*;

import util.Functional;
import util.Builders.Builder;

import static output.java.State.*;

/**
 * An object for building source code, analogous to {@link StringBuilder} for
 * building strings or the various builders in {@link util.Builders} for
 * bulding collections.
 */
public final class CodeBuilder implements Builder<String> {
    /**
     * The initial state with which this code builder was constructed.
     * By default, a code builder builds an entire compilation unit; this
     * corresponds to an initial state of {@link State#TOPLEVEL}.
     * @see State
     */
    public final State initialState;
    
    private final LinkedList<Scope> scopeStack;

    private final SortedSet<ImportStatement> imports;

    private String pkgStmt = "";
    
    /**
     * Default constructor. Sets up the code builder to build an entire
     * compilation unit (i.e. .java file).
     */
    public CodeBuilder() {
        this(TOPLEVEL);
    }

    /**
     * Construct a code builder to be embedded in another. For instance, if
     * you want to add a method to a class without passing the code builder
     * for the class, you can have the other method construct its own with the
     * initial state {@link State#CLASS}.
     * @param initialState The initial state for this code builder.
     * @see State
     */
    public CodeBuilder(State initialState) {
        this.initialState = initialState;
        scopeStack = new LinkedList<Scope>();
        imports = new TreeSet<ImportStatement>();

        scopeStack.add(new Scope(initialState)); 
    }

    /**
     * Get the current state of the code builder. This corresponds to the type
     * of the innermost scope still being built - in other words, this is where
     * the next line passed to {@link #append(String)} will go.
     * @return The current state.
     */
    public State getState() {
        return scopeStack.getFirst().state;
    }
    
    private CodeBuilder addImport(ImportStatement newStmt) {
        if (imports.contains(newStmt))
            return this;
        
        if (newStmt.isWildcard) {
            // Check for matching statements already given
            final ImportStatement succ = newStmt.successor();
            final SortedSet<ImportStatement> subset =
                imports.subSet(newStmt, succ);
            for (Iterator<ImportStatement> iter = subset.iterator();
                    iter.hasNext();)
                if (newStmt.matches(iter.next()))
                    iter.remove();
        } else // Make sure we don't already have a matching statement
            if (imports.contains(newStmt.matchingWildcard()))
                return this;
        
        imports.add(newStmt);
        return this;
    }
    
    private CodeBuilder addImport(String target, boolean isStatic) {
        return addImport(new ImportStatement(target, isStatic));
    }
    
    /**
     * Add an import statement to the top level. This will work even if this
     * builder is not building the top level (that is, {@link #initialState}
     * is not {@link State#TOPLEVEL}), so long as it is passed directly to
     * the {@link #append(CodeBuilder)} method of a top-level code builder (or
     * a builder that is so passed, and so forth). 
     * @param target The name of the package to import. The same as would be
     *      found in the statement "import stuff.package;", but without the
     *      "import" or the semicolon. The name can end in an asterisk, just as
     *      in the import statement.
     * @return {@code this}, for chaining.
     */
    public CodeBuilder addImport(String target) {
        return addImport(target, false);
    }
    
    /**
     * Call {@code addImport(String.format(fmt, args))}.
     * @see #addImport(String)
     */
    public CodeBuilder addImport(String fmt, Object ... args) {
        return addImport(String.format(fmt, args));
    }
    
    /**
     * Add a static import statement to the top level.
     * @param target The identifier of the class member to import.
     * @return {@code this}, for chaining.
     * @see #addImport(String)
     */
    public CodeBuilder addStaticImport(String target) {
        return addImport(target, true);
    }
    
    /**
     * Call {@code addStaticImport(String.format(fmt, args))}.
     * @see #addStaticImport(String)
     */
    public CodeBuilder addStaticImport(String fmt, Object ... args) {
        return addStaticImport(String.format(fmt, args));
    }
    
    /**
     * Add a package declaration. Can only be used in the
     * {@link State#TOPLEVEL} state, and only once.
     * @param pkgName The name of the package to declare.
     * @return {@code this}, for chaining.
     */
    public CodeBuilder addPackageStatement(String pkgName) {
        if (getState() != TOPLEVEL)
            throw new CodeBuilderStateError(
                    "can only specify package at top level");
        if (!pkgStmt.equals(""))
            throw new CodeBuilderStateError(
                    "can only give one package statement");
        
        pkgStmt = String.format("package %s;\n", pkgName);
        return this;
    }
    
    /**
     * Call {@code addPackageStatement(String.format(fmt, args))}.
     * @see #addPackageStatement(String)
     */
    public CodeBuilder addPackageStatement(String fmt, Object ... args) {
        return addPackageStatement(String.format(fmt, args));
    }
    
    private CodeBuilder beginScope(State state) {
        if (!getState().canContain(CLASS))
            throw new CodeBuilderStateError(String.format(
                    "%s scope cannot contain %s", getState(), state));
        final Scope newScope = new Scope(state);
        scopeStack.addFirst(newScope);
        return this;
    }
    
    private CodeBuilder finishScope(State state) {
        if (getState() != state)
            throw new CodeBuilderStateError(String.format(
                    "attempt to end %s scope while actually in %s",
                    state, getState()
                    ));
        final Scope scope = scopeStack.removeFirst();
        scopeStack.getFirst().append(scope);
        return this;
    }
    
    /**
     * Add to the code being built.
     * @param code The code to add. Must include the semicolon if the
     *      code is a statement; however, a newline will be added
     *      automatically. (If the code spans more than one line, the line
     *      breaks will be respected.) The newline will be omitted if code is
     *      empty (for a blank line use {@code append("\n")}).
     * @return {@code this}, for chaining.
     */
    public CodeBuilder append(String code) {
        if (!code.equals(""))
            scopeStack.getFirst().append(code);
        return this;
    }
    
    /**
     * Call {@code append(String.format(fmt, args))}.
     * @see #append(String)
     */
    public CodeBuilder append(String fmt, Object ... args) {
        return append(String.format(fmt, args));
    }
    
    /**
     * Absorb the contents of another code builder. This code builder
     * <i>must</i> be complete - that is, all enclosed scopes must be finished
     * and the builder at its initial state. Furthermore, this initial state
     * must be the same as this builder's current state.
     * @param builder The builder with the code to append.
     * @return {@code this}, for chaining.
     * @throws CodeBuilderStateError If the current state does not match the
     *      initial state of the other builder or it is not complete.  
     */
    public CodeBuilder append(CodeBuilder builder) {
        if (builder.initialState != getState())
            throw new CodeBuilderStateError(String.format(
                    "attempting to append builder for state %s in state %s",
                            builder.initialState, getState()));
        append(builder.finish());
        for (ImportStatement stmt : builder.imports)
            addImport(stmt);
        return this;
    }
    
    /**
     * Add to the code being built, using
     * {@link String#format(String, Object[])}.
     * {@code builder.format(fmt, arg1, arg2)} is shorthand for
     * {@code builder.append(String.format(fmt, arg1, arg2))} and does the same
     * thing.
     * @param fmt The format string.
     * @param args The arguments referred to in the format string.
     * @return {@code this}, for chaining.
     * @see String#format(String, Object[])
     */
    public CodeBuilder format(String fmt, Object ... args) {
        return append(String.format(fmt, args));
    }
    
    /**
     * Begin building a class (or interface or enum). The state will become
     * {@link State#CLASS}.
     * @param decl The declaration for the class. This means anything that
     *      would be in the class statement, up to but <i>not</i> including
     *      the opening brace. 
     * @return {@code this}, for chaining.
     */
    public CodeBuilder beginClass(String decl) {
        return append(decl + " {").beginScope(CLASS);
    }
    
    /**
     * Call {@code beginClass(String.format(fmt, args))}.
     * @see #beginClass(String)
     */
    public CodeBuilder beginClass(String fmt, Object ... args) {
        return beginClass(String.format(fmt, args));
    }
    
    /**
     * Finish building a class (or interface or enum).
     * @return {@code this}, for chaining.
     * @throws CodeBuilderStateError If the current state is not
     *      {@link State#CLASS}.
     */
    public CodeBuilder finishClass() {
        return finishScope(CLASS).append("}");
    }
    
    /**
     * Begin building a method (or constructor). The state will become
     * {@link State#METHOD}.
     * @param decl The declaration for the method. This means anything that
     *      would be in the method declaration in the code, up to but
     *      <i>not</i> including the opening brace.
     * @return {@code this}, for chaining.
     * @throws CodeBuilderStateError If the current state is not
     *      {@link State#CLASS}.
     */
    public CodeBuilder beginMethod(String decl) {
        return append(decl + " {").beginScope(METHOD);
    }
    
    /**
     * Call {@code beginMethod(String.format(fmt, args))}.
     * @see #beginMethod(String)
     */
    public CodeBuilder beginMethod(String fmt, Object ... args) {
        return beginMethod(String.format(fmt, args));
    }
    
    /**
     * Finish building a method (or constructor).
     * @return {@code this}, for chaining.
     * @throws CodeBuilderStateError If the current state is not
     *      {@link State#METHOD}.
     */
    public CodeBuilder finishMethod() {
        return finishScope(METHOD).append("}");
    }
    
    /**
     * Begin building a static block for a class. The state will become
     * {@link State#METHOD}.
     * @return {@code this}, for chaining.
     * @throws CodeBuilderStateError If the current state is not
     *      {@link State#CLASS}.
     */
    public CodeBuilder beginStaticBlock() {
        return append("static {").beginScope(METHOD);
    }
    
    /**
     * Finish building a static block for a class.
     * @return {@code this}, for chaining.
     * @throws CodeBuilderStateError If the current state is not
     *      {@link State#METHOD}.
     */public CodeBuilder finishStaticBlock() {
        return finishMethod();
    }
    
    /**
     * Begin building some kind of code block within a method. This could be
     * anything involving braces - an {@code if} statement, a {@code for} loop,
     * a {@code switch} block, even a bare <code>{ ... }</code> block used for
     * scoping. The one thing it <i>shouldn't</i> be is an inner class
     * declaration - for that, use {@link #beginClass(String)}.
     * <p>
     * The current state will become {@link State#BLOCK}.
     * @param header The part of the statement before the opening brace -
     * for instance {@code if (condition)}, {@code switch(thing)}, {@code do},
     * or the empty string for bare braces. May include a label declaration
     * ({@code LABEL:}) if you're into that sort of thing.
     * @return {@code this}, for chaining.
     * @throws CodeBuilderStateError If the current state is neither
     *      {@link State#METHOD} nor {@link State#BLOCK}.
     */
    public CodeBuilder beginBlock(String header) {
        final String maybeSpace = header.equals("") ? "" : " ";
        return append(header + maybeSpace + "{").beginScope(BLOCK);
    }
    
    /**
     * Call {@code beginBlock(String.format(fmt, args))}.
     * @see #beginBlock(String)
     */
    public CodeBuilder beginBlock(String fmt, Object ... args) {
        return beginBlock(String.format(fmt, args));
    }
    
    /**
     * Finish building a code block.
     * @return {@code this}, for chaining.
     * @throws CodeBuilderStateError If the current state is not
     *      {@link State#BLOCK}.
     */
    public CodeBuilder finishBlock() {
        return finishScope(BLOCK).append("}");
    }
    
    /**
     * Finish building a code block requiring further code after the closing
     * brace. For instance, this could be the {@code while} part of a do/while
     * loop.
     * @param afterBrace The code to go after the brace, including any necessary
     *      semicolon.
     * @return {@code this}, for chaining.
     * @throws CodeBuilderStateError If the current state is not
     *      {@link State#BLOCK}.
     */
    public CodeBuilder finishBlock(String afterBrace) {
        return finishScope(BLOCK).append("} " + afterBrace);
    }
    
    /**
     * Call {@code finishBlock(String.format(fmt, args))}.
     * @see #finishBlock(String)
     */
    public CodeBuilder finishBlock(String fmt, Object ... args) {
        return finishBlock(String.format(fmt, args));
    }
    
    /**
     * Finish building a code block, but begin another immediately. Used for
     * the sort of construction often seen in chained if/else if/else code:
     * <p><blockquote><pre>
     * if (condition) {
     *     // ...
     * } else if (otherCondition) {
     *     // ...
     * } else if (anotherCondition) {
     *     // ...
     * } else {
     *     // ...
     * }</pre>
     * </blockquote><p>
     * Each of the {@code else}-{@code if}s and the {@code else} would be
     * generated by a call to this method. (The original {@code if} and the
     * final closing brace would be generated by
     * {@code beginBlock("if (condition)")} and {@code finishBlock()},
     * respectively.)
     * @param betweenBraces The code to be put between the closing brace of the
     *      old block and the opening brace of the new one. (For example, the
     *      "{@code else if}" or "{@code else}" above.) 
     * @return {@code this}, for chaining.
     * @throws CodeBuilderStateError If the current state is not
     *      {@link State#BLOCK}.
     */
    public CodeBuilder chainBlock(String betweenBraces) {
        return finishScope(BLOCK)
                .append("} " + betweenBraces + " {")
                .beginScope(BLOCK);
    }
    
    /**
     * Call {@code chainBlock(String.format(fmt, args))}.
     * @see #chainBlock(String)
     */
    public CodeBuilder chainBlock(String fmt, Object ... args) {
        return chainBlock(String.format(fmt, args));
    }
    
    /**
     * Add a one-line-style comment to the code being built. The given string
     * may actually span multiple lines; each will be prefixed appropriately.
     * @param comment The text of the comment.
     * @return {@code this}, for chaining.
     */
    public CodeBuilder appendComment(String comment) {
        for (String line : comment.split("\\n"))
            append("// " + line);
        return this;
    }
    
    /**
     * Call {@code appendComment(String.format(fmt, args))}.
     * @see #appendComment(String)
     */
    public CodeBuilder appendComment(String fmt, Object ... args) {
        return appendComment(String.format(fmt, args));
    }
    
    /**
     * Begin building a multi-line comment. The state will become
     * {@link State#COMMENT}.
     * @return {@code this}, for chaining.
     * @throws CodeBuilderStateError If the current state is already
     *         {@link State#COMMENT}.
     */
    public CodeBuilder beginComment() {
        return append("/*")
                .beginScope(COMMENT);
    }
    
    /**
     * Begin building a JavaDoc comment. The state will become
     * {@link State#COMMENT}.
     * @return {@code this}, for chaining.
     * @throws CodeBuilderStateError If the current state is already
     *         {@link State#COMMENT}.
     */
    public CodeBuilder beginJavaDoc() {
        return append("/**")
                .beginScope(COMMENT);
    }
    
    /**
     * Finish building a multi-line comment. This includes JavaDoc comments.
     * @return {@code this}, for chaining.
     * @throws CodeBuilderStateError If the current state is not
     *         {@link State#COMMENT}.
     */
    public CodeBuilder finishComment() {
        return finishScope(COMMENT)
                .append("*/");
    }
    
    /**
     * Return the code generated by the builder so far. If there remain open
     * scopes, this code will be incomplete (there will be unmatched opening
     * braces).
     * @return The generated code.
     */
    public String toString() {
        StringBuilder builder = new StringBuilder();
        if (pkgStmt.length() > 0)
            builder.append(pkgStmt + '\n');
        if (!imports.isEmpty())
            builder.append(Functional.join(imports, "\n") + "\n\n");
        builder.append(scopeStack.getFirst());
        if (builder.length() > 0)
            builder.append('\n');
        return builder.toString();
    }
    
    /**
     * Return the code generated by the builder. It must be complete; that is,
     * each call to a begin...() method must have been matched by one to the
     * corresponding end...() method.
     * @return The generated code.
     * @throws CodeBuilderStateError If there remain unfinished scopes.
     */
    public String finish() {
        if (scopeStack.size() != 1)
            throw new CodeBuilderStateError(String.format(
                    "attempting to finish building when %s block still open",
                    getState()));
        return toString();
    }

    @SuppressWarnings("unused")
    private final class CodeBuilderStateError extends RuntimeException {
        private CodeBuilderStateError() {
            super(CodeBuilder.this.pkgStmt);
        }
        
        private CodeBuilderStateError(String message) {
            super(message);
        }
    
        private CodeBuilderStateError(String message, Throwable cause) {
            super(message, cause);
        }
    
        private CodeBuilderStateError(Throwable cause) {
            super(getDefaultMsg(), cause);
        }
    }

    private String getDefaultMsg() {
        return String.format(
                "code builder in invalid state %s for operation", getState());
    }
}
