package parser;

import java.util.*;
import java.util.regex.*;

import util.Functional;
import util.Pair;
import util.Predicate;

/** Node implementation for parsing GUI files. */
class GUINode
        implements Node, ParserTreeConstants, Iterable<GUINode> {
    /** The parent of this node */
    protected Node parent = null;
    /** The children of this node */
    protected List<GUINode> children = new LinkedList<GUINode>();
    /** The first token of this node. This is only kept track of until we
     *  find the last token, at which point the <code>tokens</code> list is
     *  created. */
    private Token firstToken = null;
    /** The tokens comprising this node. From these, the original text can be
     *  (mostly) reconstructed, and more importantly the line numbers can be
     *  found. */
    protected List<Token> tokens = new LinkedList<Token>();
    /** The kind of token, as specified in the grammar. */
    final Kind kind;
    /** The parser which generated the node. */
    final Parser parser;

    /* @see parser.Node#jjtOpen() */
    public void jjtOpen() {}
    /* @see parser.Node#jjtClose() */
    public void jjtClose() { children.removeAll(Collections.singleton(null)); }
    /* @see parser.Node#jjtGetParent() */
    public Node jjtGetParent() { return parent; }
    /* @see parser.Node#jjtSetParent(parser.Node) */
    public void jjtSetParent(Node parent) { this.parent = parent; }
    /* @see parser.Node#jjtGetChild(int) */
    public GUINode jjtGetChild(int ix) { return children.get(ix); }
    /* @see parser.Node#jjtGetNumChildren() */
    public int jjtGetNumChildren() { return children.size(); }

    /* @see parser.Node#jjtAddChild(parser.Node, int) */
    public void jjtAddChild(Node child, int ix) {
        while (ix > children.size())
            children.add(null);
        children.add(ix, (GUINode) child);
    }
    
    /** @param kind The kind of node, as specified in the grammar */
    GUINode(Parser parser, Kind kind) {
        this.parser = parser;
        this.kind = kind;
    }

    /** @return <code>true</code> if this node is an accumulation statement. */
    boolean isAccumStatement() { return kind == Kind.ACCUMSTATEMENT; }
    /** @return <code>true</code> if this node is an attribute statement. */
    boolean isAttrStatement() { return kind == Kind.ATTRSTATEMENT; }
    /** @return <code>true</code> if this node is bare string. */
    boolean isBareString() { return kind == Kind.BARESTRING; }
    /** @return <code>true</code> if this node is the whole file. */
    boolean isGUIFile() { return kind == Kind.GUIFILE; }
    /** @return <code>true</code> if this node is an identifier. */
    boolean isIdentifier() { return kind == Kind.IDENTIFIER; }
    /** @return <code>true</code> if this node is a quoted string. */
    boolean isStringValue() { return kind == Kind.STRINGVALUE; }
    /** @return <code>true</code> if this node is of void type. */
    boolean isVoid() { return kind == Kind.VOID; }
    /** @return <code>true</code> if this node is a declarator. */
    boolean isWidgetDeclarator() { return kind==Kind.WIDGETDECLARATOR; }
    /** @return <code>true</code> if this node is a declaration statement. */
    boolean isWidgetStatement() { return kind == Kind.WIDGETSTATEMENT; }
    
    /** @return <code>true</code> if this is any kind of statement */
    boolean isStatement() {
        return isWidgetStatement() || isAccumStatement() || isAttrStatement(); 
    }
    
    /** @return <code>true</code> if this is either kind of string literal */
    boolean isString() {
        return isStringValue() || isBareString();
    }
    
    /** @return <code>true</code> if this is a declarator or identifier */
    boolean isDeclaration() {
        return isWidgetDeclarator() || isIdentifier();
    }
    
    /** Split a list of nodes into a list of declarations and a list of
     *  statements. Used to identify which parts of a statement are the
     *  declarators/identifiers, and which parts are the block.
     *  
     *  @param nodes A list of nodes to separate.
     *  @return The list of declarations, paired with the list of statements */
    static Pair<? extends List<GUINode>, ? extends List<GUINode>>
            splitDeclsFromStmts(List<GUINode> nodes) {
                
        return Functional.span(nodes, new Predicate<GUINode>() {
            public boolean f(GUINode node) { return !node.isStatement(); }
        }, new LinkedList<GUINode>(), new LinkedList<GUINode>());
    }
    
    /** @param token The token to be remembered as this node's first */
    void setFirstToken(Token token) {
        firstToken = token;
    }

    /** @param lastToken The token to be remembered as this node's last 
     * 
     *  @throws NullPointerException if setFirstToken() has not been called */
    void setLastToken(Token lastToken) {
        for (Token token = firstToken; token != lastToken; token = token.next)
            tokens.add(token);

        // Add the last token
        tokens.add(lastToken);
    }

    private static final int INDENTWIDTH = 2;
    private String mkSpaces(int count) {
        char chars[] = new char[count];
        Arrays.fill(chars, ' ');
        return new String(chars);
    }
    
    private String display(int indent) {
        StringBuilder builder = 
            new StringBuilder(mkSpaces(indent));

        builder.append(kind.toString() + ": ");
        
        // If we have no children, output the tokens
        if (jjtGetNumChildren() == 0) {
            builder.append('\u201c');
            String tokString = getText();
            builder.append(tokString == null ? "<null>" : tokString);
            builder.append("\u201d\n");
        }
        // Else output the children
        else {
            builder.append("\n");
            for (GUINode child : children) {
                builder.append(child == null ?
                        mkSpaces(indent + INDENTWIDTH) + "<null>" : 
                        child.display(indent + INDENTWIDTH));
            }
        }

        return builder.toString();
    }

    /** Pretty-print this node and all its children, for diagnostic purposes.
     *  @return A multi-line string representing the parse tree rooted at this
     *      node.*/
    String display() {
        return display(0);
    }
    
    public String toString() {
        switch (kind) {
        case GUIFILE:
        case VOID:
            return kind.nontermId;
        case IDENTIFIER:
        case BARESTRING:
        case STRINGVALUE:
            return String.format("%s: \u201c%s\u201d",
                    kind.nontermId, getText());
        case ATTRSTATEMENT:
            return String.format("%s: \u201c%s\u201d",
                    getChildren().get(0).getText());
        case ACCUMSTATEMENT: {
            List<GUINode> decls = splitDeclsFromStmts(getChildren()).a;
            StringBuilder builder = new StringBuilder(
                    String.format("\u201c%s\u201d", decls.get(0).getText()));
            for (GUINode decl : decls)
                builder.append(String.format(", \u201c%s\u201d",
                        decl.getText()));
            return String.format("%s: %s", kind.nontermId, builder);
        } case WIDGETDECLARATOR: {
            final GUINode first = getChildren().get(0);
            final String str = first.getText();
            if (jjtGetNumChildren() == 1) {
                switch(first.kind) {
                case IDENTIFIER:
                    return String.format("\u201c%s\u201d", str);
                case STRINGVALUE:
                case BARESTRING:
                    if (str.split("\\s").length <= 1)
                        return String.format("\u201c\"%s\"\u201d", str);
                    else
                        return String.format("\u201c\"%s ...\"\u201d",
                                str.split("\\s")[0]);
                default:
                    /* Never reached */ return null;
                }
            } else
                return String.format("\u201c%s\u201d", str);
        } case WIDGETSTATEMENT: {
            final List<GUINode> children = getChildren();
            final String type = children.get(0).getText();
            final List<GUINode> decls = splitDeclsFromStmts(
                    children.subList(1, children.size())).a;
            final String declStr = decls.isEmpty() ? 
                    "<anon>" : Functional.join(decls, ", ");
            return String.format("%s: %s %s", kind.nontermId, type, declStr);
        } default:
            /* Never reached */ return null;
        }
    }

    private String getTokenString() {
        StringBuilder builder = new StringBuilder();
        for (Token token : tokens)
            builder.append(token);
        return builder.toString();
    }
    
    private static final Pattern continuation = Pattern.compile("\\\\\\n\\s*");
    /** @return The text contained in this node, if it is a terminal symbol.
     *  @throws IllegalArgumentException If the node is not a terminal
     *      symbol. */
    String getText() {
        switch (kind) {
            case IDENTIFIER:
            case BARESTRING:
                return getTokenString();
            
            case STRINGVALUE:
                // First, get rid of quotation marks
                String ans = getTokenString();
                ans = ans.substring(1, ans.length() - 1);

                ans = continuation.matcher(ans).replaceAll("");

                return ans;
                
            default:
                throw new IllegalArgumentException("not a terminal symbol");
        }
    }

    /** @return An unmodifiable view of the tokens comprising this node */
    List<Token> getTokens() {
        return Collections.unmodifiableList(tokens);
    }

    /** @return The line number on which the node starts 
     *  @throws NullPointerException If setFirstToken() has not been called */
    int getLineNum() {
        return tokens.get(0).beginLine;
    }
    
    /** Iterates through the children of this node.
     *  @see java.lang.Iterable#iterator() */
    public Iterator<GUINode> iterator() {
        return getChildren().iterator();
    }

    /** @return An unmodifiable view of the child nodes of this node */
    List<GUINode> getChildren() {
        return Collections.unmodifiableList(children);
    }
}
