package types;

import java.util.*;

import errors.*;

/** Passed to functions that may report errors in the .gui file. This way many
 *  can be collected and reported, rather than just the first exception
 *  thrown. */
public interface ErrorHandler {
    /** Handle the error. It is up to the implementor what to do with the
     * error; possibilities include printing it to the screen, adding it to a
     * list to be retrieved later, ignoring it, etc.
     * @param error The error to handle.
     * @return <code>true</code> if the error was handled. Implementors may use
     *      this to be selective about which errors they handle. */
    boolean handleError(GUIFileException error);
    
    /** @return <code>true</code> if this handler has handled any errors. */
    boolean anyErrors();
    
    /** An error handler that handles nothing. */    
    class Trivial implements ErrorHandler {
        /* @see types.ErrorHandler#handleError(errors.GUIFileException) */
        public boolean handleError(GUIFileException error) {
            return false;
        }

        /* @see types.ErrorHandler#anyErrors() */
        public boolean anyErrors() {
            return false;
        }
    }
    
    /** An error handler that handles anything, spitting it to stderr. */
    class StdErr implements ErrorHandler {
        /** {@code true} if handleError has been called at least once */
        protected boolean seenAny = false;
        
        /* @see types.ErrorHandler#handleError(errors.GUIFileException) */
        public boolean handleError(GUIFileException error) {
            System.err.println(error.getMessage());
            seenAny = true;
            return true;
        }

        /* @see types.ErrorHandler#anyErrors() */
        public boolean anyErrors() {
            return seenAny;
        }
    }
    
    /** An error handler that just stuffs everything in a list */
    class Lister implements ErrorHandler {
        private final List<GUIFileException> errs =
            new ArrayList<GUIFileException>();
        
        /** A list of all errors reported to the error handler. An immutable
         *  view. */
        public final List<GUIFileException> errors =
            Collections.unmodifiableList(errs);
            
        /* @see types.ErrorHandler#handleError(errors.GUIFileException) */
        public boolean handleError(GUIFileException error) {
            errs.add(error);
            return true;
        }

        /* @see types.ErrorHandler#anyErrors() */
        public boolean anyErrors() {
            return !errs.isEmpty();
        }
    }
}
