package types;

/**
 * Defines the event handler types allowed in %controller blocks in .gui files,
 * along with the signatures of the methods generated in the output Controller
 * interface. 
 */
public enum EventHandlerDef {
    /**
     * An event handler which is passed no additional information.
     */
    Action("Component widget, ActionEvent event",
            "java.awt.Component",
            "java.awt.event.ActionEvent"),
    
    /**
     * A listener handler for a Tree.
     */
    Tree("Component widget, TreeSelectionEvent event",
            "java.awt.Component",
            "java.awt.event.ActionEvent",
            "javax.swing.event.TreeSelectionEvent")
    ;
    
    /**
     * The signature of the generated event handling method, without the return
     * type or the name of the method. In other words, the argument list,
     * without the parentheses.
     * <p>
     * For instance, an {@link #Action} has a {@code handlerSig} of
     * {@code "Component widget"}, so the controller code generated for a
     * declaration like this: <blockquote>
     * 
     * <pre>
     * %controller {
     *     ActionHandler frobButtonClicked;
     * }
     * </pre>
     * 
     * </blockquote> ... might look like this: <blockquote>
     * 
     * <pre>
     * public interface Controller {
     *     // ...
     *     void frobButtonClicked(Component widget);
     *     // ...
     * }
     * </pre>
     * 
     * </blockquote>
     */
    public final String handlerSig;
    
    /**
     * The classes that need to be imported for use in the handler signature.
     */
    public final String[] imports;
    
    private EventHandlerDef(String handlerSig, String ... imports) {
        this.handlerSig = handlerSig;
        this.imports = imports;
    }
    
    private EventHandlerDef() {
        this("");
    }
}
