package types;
/**
 *   GUIAttributes is the implementation of TreeMap that automates
 * defaulting so that Widget definitions will not need to distinguish
 * between setting a specified property and setting the default one.
 */


import java.util.*;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import util.Builders;
import util.Functional;
import util.Pair;

import static util.Functional.*;

import errors.BadAttributeException;
import errors.BadAttributeNameException;
import errors.BadAttributeValueException;

/** A specialized extension of TreeMap which helps in handling widget
 *  attributes. Specifically, a GUIAttributes object holds a map with default
 *  values for all attributes; has methods to assist in attribute retrieval
 *  and parsing; remembers what widget it is associated with; and automates
 *  error handling by generating exceptions that, 90% of the time, should be
 *  sufficient such that a widget definition doesn't need any explicit error
 *  checking. */
public final class GUIAttributes extends TreeMap <String, List<String>> {
	
    /**
     * A map of the names of the recognized attributes to their default
     * values.
     */
    public final Map<String,List<String>> defaults;
    /**
     * A map of synonymous attribute names to their canonical equivalents.
     */
    public final Map<String, String> synonyms;
    /**
     * A map on multisynonymous attribute names to the names of their
     * component attributes.
     */
    public final Map<String, List<String>> multisynonyms;
	/**
	 * The {@link Widget} to which this object belongs.
	 */
	public final Widget widget;

    /**
     * Construct a GUIAttributes map, given a map of the defaults, a map
     * describing synonymies of the attributes, and the Widget which is to own
     * the map.
     * @param defaults A map of the names of the recognized attributes to their
     *        default values.
     * @param synonyms A map, each entry of which has an attribute name as the
     *        key and the real attribute with which it is synonymous as the
     *        value. 
     * @param multisynonyms A map, each entry of which has an attribute name as
     *        the key and the list of real attributes which it maps to as the
     *        value. For instance, if giving the "size" attribute a value of
     *        two integers is equivalent to giving the "width" and "height"
     *        attributes separately, the map would have an entry with key
     *        "size" and values "width" and "height".
     * @param widget The {@link Widget} to which this object belongs.
     */
    public GUIAttributes(Map<String,List<String>> defaults,
            Map<String, String> synonyms,
            Map<String, List<String>> multisynonyms, Widget widget) {
        this.defaults = defaults;
        this.synonyms = synonyms;
        this.multisynonyms = multisynonyms;
        this.widget = widget;
    }
    
    /**
     * Construct a GUIAttributes map, given a map of the defaults, a map
     * describing synonymies of the attributes, and the Widget which is to own
     * the map.
     * @param defaults A map of the names of the recognized attributes to their
     *        default values.
     * @param synonyms A map, each entry of which has an attribute name as the
     *        key and the real attribute with which it is synonymous as the
     *        value. 
     * @param widget The {@link Widget} to which this object belongs.
     */
    public GUIAttributes(Map<String,List<String>> defaults,
            Map<String, String> synonyms, Widget widget) {
        this(defaults, synonyms, 
                Collections.<String, List<String>> emptyMap(), widget);
    }
    
    /**
     * Construct a GUIAttributes map, given a map of the defaults and the Widget
     * which is to own it. The map will have no synonymous attributes (see
     * {@link #GUIAttributes(Map, Map, Widget)} for this).
     * @param defaults A map of the names of the recognized attributes to their
     *        default values.
     * @param widget The {@link Widget} to which this object belongs.
     */
    public GUIAttributes(Map<String,List<String>> defaults, Widget widget) {
        this(defaults, Collections.<String, String> emptyMap(), widget);
    }
    
    private String resolveMonosynonym(String attr) {
        while (synonyms.containsKey(attr))
            attr = synonyms.get(attr);
        return attr;
    }
    
    private List<String> resolveSynonym(String attr) {
        boolean isMonosynonym;
        while ((isMonosynonym = synonyms.containsKey(attr)) ||
                multisynonyms.containsKey(attr)) {
            if (isMonosynonym)
                attr = synonyms.get(attr);
            else {
                final List<String> realKeys = multisynonyms.get(attr);
                final List<String> ans = 
                    new ArrayList<String>(realKeys.size());
                for (String key : realKeys)
                    ans.add(resolveMonosynonym(key));
                return ans;
            }
        }
        return Collections.singletonList(attr);
    }
    
	/**
     * @param attrs A map of attributes, each of which will be added.
     * @throws BadAttributeNameException If any of the given attribute names
     *         are not valid for this widget.
     * @throws BadAttributeValueException If a multisynonymous attribute (see
     *         {@link WidgetDef#getMultisynonymousAttributes()} has the wrong
     *         number of values.
     */
	public void putAllAttributes(
	            Map<String, ? extends List<String>> attrs)
	        throws BadAttributeException {
	    /*
	     * Checks to make sure all the properties being added are present in the
	     * default set, in which case they are valid properties. Also resolves
         * synonymous attributes.
	     */
        
        final Map<String, List<String>> realAttrs =
            new TreeMap<String, List<String>>(attrs);
	    
	    for (String key : attrs.keySet()) {
            final String origKey = key;
            final List<String> realKeys = resolveSynonym(key);
            if (realKeys.size() == 1) {
                String realKey = realKeys.get(0);
                realAttrs.remove(key);
                realAttrs.put(realKey, attrs.get(key));
            } else {
                final List<String> values = attrs.get(key);
                
                if (realKeys.size() != values.size())
                    throw new BadAttributeValueException(
                            String.format("exactly %d values required", 
                                    realKeys.size()), widget, key, values);
                
                realAttrs.remove(key);
                for (Pair<String, String> k_v : 
                        Functional.zip(realKeys, values))
                    realAttrs.put(k_v.a, Collections.singletonList(k_v.b));                
            }
            
            for (String realKey : realKeys)
                if (!defaults.containsKey(realKey))
                    throw new BadAttributeNameException(widget, origKey);
        }
	    
	    super.putAll(realAttrs);
	}
	
    private List<String> putExactlyThisAttribute(
            String attr, String ... values) throws BadAttributeNameException {
        if (!defaults.containsKey(attr))
            throw new BadAttributeNameException(widget, attr);
        return super.put(attr, Arrays.asList(values));
    }
    
    /**
     * Add an attribute, with arbitrarily many values, to this map.
     * @param attr The name of the attribute.
     * @param values Zero or more values for the attribute.
     * @return The previous values of the attribute, or {@code null} if it was
     *      at its default setting.
     * @throws BadAttributeNameException If the attribute name is not 
     *      recognized (see {@link #defaults}).
     * @throws BadAttributeValueException If the attribute is a multisynonym
     *      (see {@link WidgetDef#getMultisynonymousAttributes()}) with the
     *      wrong number of values.
     */
    public List<String> putAttribute(String attr, String ... values) 
            throws BadAttributeException {
        
        final List<String> realKeys = resolveSynonym(attr);
        if (realKeys.size() == 1)
            return putExactlyThisAttribute(realKeys.get(0));
        else {
            if (realKeys.size() != values.length)
                throw new BadAttributeValueException(
                        String.format("exactly %d values required", 
                                realKeys.size()), widget, attr, 
                                Arrays.asList(values));
            
            final List<String> ans = new ArrayList<String>(values.length); 
            
            for (Pair<String, String> k_v :
                    Functional.zip(realKeys, Arrays.asList(values)))
                ans.add(putExactlyThisAttribute(k_v.a, k_v.b).get(0));
            
            return ans;
        }
    }

    /**
     * Add an attribute, with arbitrarily many values, to this map.
     * @param attr The name of the attribute.
     * @param values A list of values for the attribute.
     * @return The previous values of the attribute, or {@code null} if it was
     *      at its default value.
     * @throws BadAttributeNameException If the attribute name is not 
     *      recognized (see {@link #defaults}).
     */
    public List<String> putAttribute(String attr, List<String> values) 
            throws BadAttributeException {
        return putAttribute(attr, values.toArray(new String[values.size()]));
    }

    /**
     * Add an attribute to this map.
     * @param attr The name of the attribute.
     * @param values A list of its values.
     * @return The previous values of the attribute, or {@code null} if it was
     *      at its default setting.
     * @throws IllegalArgumentException If the attribute name is not
     *      recognized (see {@link #defaults}).
     * @deprecated Use {@link #putAttribute(String, String[])} when possible,
     *      as it throws the proper checked exception,
     *      {@link BadAttributeNameException}, on error. This method is only
     *      provided for compatibility with the base class.
     */
    @Deprecated @Override
    public List<String> put(String attr, List<String> values) {
        try {
            return putAttribute(attr, values);
        } catch (BadAttributeException e) {
            throw new IllegalArgumentException(e);
        }
    }
    
    /** Retrieve all values of an attribute by name, as strings.
     *  @param attr The name of the attribute.
     *  @return The values of the attribute, or <code>null</code> if there is
     *      no such attribute.
     *  @deprecated Use {@link #getAll} instead; it typechecks its argument
     *      and throws suitable exceptions on invalid attribute values. Better
     *      yet, use one of {@link #getString}, {@link #getStrings},
     *      {@link #getInteger}, etc., as needed. This method is only meant to
     *      allow GUIAttributes objects to be cast to Maps and still work
     *      (somewhat). */
    @Deprecated @Override
	public List<String> get(Object attr) {
        try {
            return getAll((String) attr);
        } catch (BadAttributeException e) {
            throw new IllegalArgumentException(e);
        }
    }
    
    private List<String> getAllOfExactlyThis(String attr) 
            throws BadAttributeNameException {
        if (super.containsKey(attr)){
            return super.get(attr);
        } else if (defaults.containsKey(attr)) {
            return defaults.get(attr);
        } else {
            throw new BadAttributeNameException(widget, attr);
        }   
    }
    
    /**
     * Retrieve all values of an attribute by name, as strings.
     * @param attr The name of the attribute.
     * @return The values of the attribute, as strings.
     * @throws BadAttributeNameException If there is no such attribute.
     */
	public List<String> getAll(String attr) throws BadAttributeException {
		final List<String> realKeys = resolveSynonym(attr);
        if (realKeys.size() == 1)
            return getAllOfExactlyThis(realKeys.get(0));
        else {
            final List<String> ans = new ArrayList<String>(realKeys.size());
            for (String key : realKeys)
                ans.add(getOneOfExactlyThis(key));
            return ans;
        }
	}

    /**
     * Retrieve all values of an attribute which allows exactly N values, for
     * some N. Checks to make sure that there are N items in the attribute
     * list.
     * @param attr The name of the atttribute.
     * @param quantity The number of values to require.
     * @return The values of the attribute.
     * @throws BadAttributeNameException If there is no such attribute.
     * @throws BadAttributeValueException If the attribute has more or fewer
     *      values than {@code quantity}.
     */
    public List<String> getMany(String attr, int quantity) 
            throws BadAttributeException {
        List<String> values = getAll(attr);
        if (values.size() != quantity) {
            throw new BadAttributeValueException(
                    "expected " + quantity + " values", widget, attr,
                    quoteAndJoin(values, ", "));
        }
        return values;
    }
    
    private String getOneOfExactlyThis(String attr) 
            throws BadAttributeException {
        List<String> values = getAllOfExactlyThis(attr);
        if (values.size() != 1) {
            throw new BadAttributeValueException(
                    "Only one value allowed", widget, attr,
                    quoteAndJoin(values, ", "));
        }
        return values.get(0);
    }
    
    /**
     * Retrieve the value of an attribute which only allows one value. Checks
     * to make sure that there is only one item in the attribute list.
     * @param attr The name of the atttribute.
     * @return The value of the attribute.
     * @throws BadAttributeNameException If there is no such attribute.
     * @throws BadAttributeValueException If the attribute has more than one
     *     value.
     */
    public String getOne(String attr) throws BadAttributeException {
        List<String> values = getAll(attr);
        if (values.size() != 1) {
            throw new BadAttributeValueException(
                    "Only one value allowed", widget, attr,
                    quoteAndJoin(values, ", "));
        }
        return values.get(0);
    }
    
    /** Retrieve all values of an attribute as strings. Equivalent to getAll().
     *  @param attr The name of the attribute.
     *  @return The values of the attribute, as strings.
     *  @throws BadAttributeNameException If there is no such attribute. */
    public List<String> getStrings(String attr) 
            throws BadAttributeException {
        return getAll(attr);
    }

    /** Retrieve, as a string, the value of an attribute which only allows one
     *      value. Equivalent to getOne().
     *  @param attr The name of the attribute.
     *  @return The value of the attribute, as a string, if it has only one
     *      value.
     *  @throws BadAttributeNameException If there is no such attribute.
     *  @throws BadAttributeValueException If the attribute has more than one
     *      value. */
    public String getString(String attr) throws BadAttributeException {
        return getOne(attr);
    }
    
    /** Retrieve all values of an attribute as integers.
     *  @param attr The name of the attribute.
     *  @return The values of the attribute, as integers.
     *  @throws BadAttributeNameException If there is no such attribute. 
     *  @throws BadAttributeValueException If any attribute value fails to
     *      parse as an integer. */
    public List<Integer> getIntegers(String attr)
            throws BadAttributeException {
        final List<String> values = getAll(attr);
        final List<Integer> ans = new ArrayList<Integer>(values.size());
        for (String value : values)
            ans.add(parseInteger(attr, value));
        return ans;
    }
    
    /** Retrieve, as an integer, the value of an attribute which only allows
     *      one value.
     *  @param attr The name of the attribute.
     *  @return The value of the attribute, as an integer, if it has only one
     *      value.
     *  @throws BadAttributeNameException If there is no such attribute.
     *  @throws BadAttributeValueException If the value fails to parse as an
     *      integer, or the attribute has more than one value. */
    public int getInteger(String attr) throws BadAttributeException {
        return parseInteger(attr, getOne(attr));
    }
    
    /**
     * Retrieve, as a pair of integers, the value of an attribute which only
     * allows exactly two values.
     * @param attr The name of the attribute.
     * @return The value of the attribute, as an integer, if it has only one
     *         value.
     * @throws BadAttributeNameException If there is no such attribute.
     * @throws BadAttributeValueException If the value fails to parse as an
     *         integer, or the attribute has more than one value.
     */
    public Pair<Integer, Integer> getIntegerPair(String attr) 
            throws BadAttributeException {
        List<Integer> list = getIntegers(attr);
        if (list.size() != 2)
            throw new BadAttributeValueException("expected a pair of integers",
                    widget, attr, getStrings(attr));
        return Pair.of(list.get(0), list.get(1));
    }
    
    /**
     * Parse a string as an integer attribute value, as {@link #getInteger}
     * would.
     * @param attr The name of the attribute, passed to the exception if one is
     *        thrown.
     * @param value The value to parse.
     * @return The attribute value, as an integer.
     * @throws BadAttributeValueException If the value fails to parse as an
     *         integer.
     */
    public int parseInteger(String attr, String value)
            throws BadAttributeValueException {
        try {
            return Integer.valueOf(value);
        } catch (NumberFormatException e) {
            throw new BadAttributeValueException(
                    "integer expected", widget, attr, value);
        }
    }
    
    /** Retrieve all values of an attribute as doubles.
     *  @param attr The name of the attribute.
     *  @return The values of the attribute, as doubles.
     *  @throws BadAttributeNameException If there is no such attribute. 
     *  @throws BadAttributeValueException If any attribute value fails to
     *      parse as a double. */
    public List<Float> getFloats(String attr)
            throws BadAttributeException {
        final List<String> values = getAll(attr);
        final List<Float> ans = new ArrayList<Float>(values.size());
        for (String value : values)
            ans.add(parseFloat(attr, value));
        return ans;
    }
    
    /** Retrieve, as a double, the value of an attribute which only allows
     *      one value.
     *  @param attr The name of the attribute.
     *  @return The value of the attribute, as a double, if it has only one
     *      value.
     *  @throws BadAttributeNameException If there is no such attribute.
     *  @throws BadAttributeValueException If the value fails to parse as a
     *      double, or the attribute has more than one value. */
    public float getFloat(String attr) throws BadAttributeException {
        return parseFloat(attr, getOne(attr));
    }
    
    /**
     * Parse a string as a float attribute value, as {@link #getFloat} would.
     * @param attr The name of the attribute, passed to the exception if one is
     *        thrown.
     * @param value The value to parse.
     * @return The attribute value, as a float.
     * @throws BadAttributeValueException If the value fails to parse as a
     *         float.
     */
    public float parseFloat(String attr, String value)
            throws BadAttributeValueException {
        try {
            return Float.valueOf(value);
        } catch (NumberFormatException e) {
            throw new BadAttributeValueException(
                    "real number expected", widget, attr, value);
        }
    }
  
    /** Retrieve all values of an attribute as doubles.
     *  @param attr The name of the attribute.
     *  @return The values of the attribute, as doubles.
     *  @throws BadAttributeNameException If there is no such attribute. 
     *  @throws BadAttributeValueException If any attribute value fails to
     *      parse as a double. */
    public List<Double> getDoubles(String attr)
            throws BadAttributeException {
        final List<String> values = getAll(attr);
        final List<Double> ans = new ArrayList<Double>(values.size());
        for (String value : values)
            ans.add(parseDouble(attr, value));
        return ans;
    }
    
    /** Retrieve, as a double, the value of an attribute which only allows
     *      one value.
     *  @param attr The name of the attribute.
     *  @return The value of the attribute, as a double, if it has only one
     *      value.
     *  @throws BadAttributeNameException If there is no such attribute.
     *  @throws BadAttributeValueException If the value fails to parse as a
     *      double, or the attribute has more than one value. */
    public double getDouble(String attr) throws BadAttributeException {
        return parseDouble(attr, getOne(attr));
    }
    
    /**
     * Parse a string as a double attribute value, as {@link #getDouble} would.
     * @param attr The name of the attribute, passed to the exception if one is
     *        thrown.
     * @param value The value to parse.
     * @return The attribute value, as a double.
     * @throws BadAttributeValueException If the value fails to parse as a
     *         double.
     */
    public double parseDouble(String attr, String value)
            throws BadAttributeValueException {
        try {
            return Double.valueOf(value);
        } catch (NumberFormatException e) {
            throw new BadAttributeValueException(
                    "real number expected", widget, attr, value);
        }
    }
  
    /** Retrieve all values of an attribute as enumeration constants.
     *  @param <E> The enumeration class. 
     *  @param attr The name of the attribute.
     *  @param cls The enumeration class.
     *  @return The values of the attribute, as doubles.
     *  @throws BadAttributeNameException If there is no such attribute. 
     *  @throws BadAttributeValueException If any attribute value fails to
     *      parse as a double. */
    public <E extends Enum> List<E> getEnums(String attr, Class<E> cls)
            throws BadAttributeException {
        final List<String> values = getAll(attr);
        final List<E> ans = new ArrayList<E>(values.size());
        for (String value : values)
            ans.add(parseEnum(attr, value, cls));
        return ans;
    }
    
    /** Retrieve, as an enumeration constant, the value of an attribute which
     *      only allows one value.
     *  @param <E> The enumeration class.
     *  @param attr The name of the attribute.
     *  @param cls The enumeration class.
     *  @return The value of the attribute, as an integer, if it has only one
     *      value.
     *  @throws BadAttributeNameException If there is no such attribute.
     *  @throws BadAttributeValueException If the value fails to parse as an
     *      integer, or the attribute has more than one value. */
    public <E extends Enum> E getEnum(String attr, Class<E> cls)
            throws BadAttributeException {
        return parseEnum(attr, getOne(attr), cls);
    }
    
    /**
     * Parse a string as an enumeration constaint, as {@link #getEnum} would.
     * @param <E> The enumeration class.
     * @param attr The name of the attribute, passed to the exception if one is
     *        thrown.
     * @param value The value to parse.
     * @param cls The enumeration class.
     * @return The attribute value, as a value of {@code <E>}.
     * @throws BadAttributeValueException If the value fails to parse as a
     *         constant of the given enumeration.
     */
    @SuppressWarnings("unchecked")
    public <E extends Enum> E parseEnum(String attr, String value,
            Class<E> cls)
                throws BadAttributeValueException {
        try {
            return (E) E.valueOf(cls, value);
        } catch (IllegalArgumentException e) {
            final E enums[] = cls.getEnumConstants();
            ArrayList<String> ids = new ArrayList<String>(enums.length);
                for (E enumConst : enums)
                    ids.add(enumConst.name());
            
            final String msg = String.format(
                    "expected one of the following: %s", 
                    join(wrapEach(ids, "\u201c", "\u201d"), ", "));
            throw new BadAttributeValueException(
                    msg, widget, attr, value);
        }
    }

    /** Retrieve all values of an attribute as booleans.
     *  @param attr The name of the attribute.
     *  @return The values of the attribute, as booleans.
     *  @throws BadAttributeNameException If there is no such attribute. 
     *  @throws BadAttributeException If any attribute value fails to parse as
     *      a boolean (see {@link #getBoolean}).
     *  @see #getBoolean */
    public List<Boolean> getBooleans(String attr)
            throws BadAttributeException {
        final List<String> values = getAll(attr);
        final List<Boolean> ans = new ArrayList<Boolean>(values.size());
        for (String value : values)
            ans.add(parseBoolean(attr, value));
        return ans;
    }
    
    /** Retrieve, as a boolean, the value of an attribute which only allows
     *      one value.
     *  @param attr The name of the attribute.
     *  @return <code>true</code> for "true", "yes", "1", or "on";
     *      <code>false</code> for "false", "no", "0", or "off". (Case
     *      insensitive.)
     *  @throws BadAttributeNameException If there is no such attribute. 
     *  @throws BadAttributeValueException If none of the above matches, or
     *      there is more than one value. */
    public boolean getBoolean(String attr) throws BadAttributeException {
        return parseBoolean(attr, getOne(attr));
    }

    /** A mapping of acceptible string values for Boolean attributes to their
     *      Boolean values. */
    public static final Map<String, Boolean> boolStrings = Builders.buildMap(
            new TreeMap<String, Boolean>(String.CASE_INSENSITIVE_ORDER))
            .putForKeys(true, "true", "yes", "1", "on")
            .putForKeys(false, "false", "no", "0", "off")
            .finish();
    
    /**
     * Parse a string as a boolean attribute value, as {@link #getBoolean} would.
     * @param attr The name of the attribute, passed to the exception if one is
     *        thrown.
     * @param value The value to parse.
     * @return The attribute value, as a boolean.
     * @throws BadAttributeValueException If the value fails to parse as a
     *         boolean.
     * @see #boolStrings
     */
    public boolean parseBoolean(String attr, String value) 
            throws BadAttributeValueException {
        final Boolean ans = boolStrings.get(value);
        if (ans != null)
            return ans;
        throw new BadAttributeValueException("Expected a Boolean value",
                widget, attr, value);
    }

    /** Retrieve all values of a color attribute as identifier strings
     *      suitable for use in generated code.
     *  @param attr The name of the attribute.
     *  @return Identifiers for the colors specified.
     *  @throws BadAttributeNameException If there is no such attribute. 
     *  @throws BadAttributeValueException If any attribute value is not
     *      recognized as a color. */
    public List<String> getColors(String attr) throws BadAttributeException {
        final List<String> values = getAll(attr);
        final List<String> ans = new ArrayList<String>(values.size());
        for (String value : values)
            ans.add(parseColor(attr, value));
        return ans;
    }
    
    /** Retrieve, as a color identifier, the value of an attribute which only
     *      allows one value. This identifier is suitable for use in
     *      generated code.
     *  @param attr The name of the attribute.
     *  @return The identifier for the color specified.
     *  @throws BadAttributeNameException If there is no such attribute. 
     *  @throws BadAttributeValueException If the value is not recognized as a
     *      color, or there is more than one value. */
    public String getColor(String attr) throws BadAttributeException {
        return parseColor(attr, getOne(attr));
    }
    
    /** The set of acceptable color names (except that colors in
     *      {@link #colorEquivs} are accepted as well). Case-insensitive. */
    public static final SortedSet<String> colors = Builders.makeCollection(
            new TreeSet<String>(String.CASE_INSENSITIVE_ORDER), 
            "black", "blue", "cyan", "gray",
            "green", "magenta", "orange", "pink", "red", "white",
            "yellow", "darkGray", "lightGray"); 
    
    /** A mapping of color synonyms to their canonical color names.
     *      Case-insensitive. */
    public static final Map<String, String> colorEquivs = Builders.buildMap(
            new TreeMap<String, String>(String.CASE_INSENSITIVE_ORDER))
            .putForKeys("darkGray",
                    "dark gray", "darkgray", "dark_gray",
                    "dark grey", "darkgrey", "dark_grey", "darkGrey")
            .putForKeys("lightGray",
                    "light gray", "lightgray", "light_gray",
                    "light grey", "lightgrey", "light_grey", "lightGrey")
            .finish();
    
    /** The regular expression matching HTML-style color values. */
    public static final Pattern htmlColorPat =
            Pattern.compile(
                    "#([0-9a-fA-F]{2})([0-9a-fA-F]{2})([0-9a-fA-F]{2})");
    
    /**
     * Parse a string as a color value, as {@link #getColor} would.
     * @param attr The name of the attribute, passed to the exception if one is
     *        thrown.
     * @param value The value to parse.
     * @return The attribute value, as a color identifier.
     * @throws BadAttributeValueException If the value fails to parse as a
     *         color.
     */
    public String parseColor(String attr, String value) 
            throws BadAttributeValueException {
        // Canonical name of value (i.e. the name of the field in Color)
        final String can;
        Matcher matcher; 
        if ((matcher = htmlColorPat.matcher(value)).matches()) {
            final String 
                r = matcher.group(1),
                g = matcher.group(2),
                b = matcher.group(3);
            return String.format("new Color(0x%s, 0x%s, 0x%s)", r, g, b);
        }
        else if (colors.contains(value))
            // Since colors is a case-insensitive set, we know only that it has
            // a member which is equivalent under case-insensitive comparison
            // to value. colors.tailSet(value).first(), then, is effectively
            // the least upper bound of value in colors; this will be the
            // actual color name, with case intact.
            can = colors.tailSet(value).first();
        else if (colorEquivs.containsKey(value))
            can = colorEquivs.get(value);
        else
            throw new BadAttributeValueException("not a recognized color",
                    widget, attr, value);
        return "Color." + can;
    }

    /**
     * Test to see if an attribute has a value (i.e. is not {@code null}).
     * @param attr The name of the attribute to test.
     * @return {@code true} if the attribute is not equal to {@code null}.
     * @throws BadAttributeException If there is no such attribute.
     */
    public boolean hasValue(String attr) throws BadAttributeException {
        return getAll(attr).get(0) != null;
    }
}
