package types;

import java.util.*;

import util.Output;

/** The GUI definition file. */
public abstract class GUIFile {
    
    /** The enumeration type of flags set in the file. */
    public static enum Flag {
        /** Indicates that public static void main() should be generated. */
        GENERATE_MAIN("generateMain"),
        /** Suppresses the generation of methods to modify attributes. */
        SUPPRESS_MODIFIERS("suppressModifiers"),
        /** Sets GENERATE_MAIN and SUPPRESS_MODIFIERS. */
        MOCKUP("mockup"),
        ;
        
        /** The keyword in the syntax corresponding to the flag. */
        public final String keyword;
        
        private static Map<String, Flag> kwdTbl = new TreeMap<String, Flag>();
        /** A mapping of keywords to associated flags. */
        public static final Map<String, Flag> keywordTable =
            Collections.unmodifiableMap(kwdTbl);
        
        static {
            for (Flag flag : Flag.values())
                kwdTbl.put(flag.keyword, flag);
        }
        
        private Flag(String keyword) {
            this.keyword = keyword;
        }
    }
    
    /** A handler declared in the GUI file as part of the controller. May be
     *  referred to in the value of onClick, etc. */
    public static class EventHandler {
        /** The type of the handler, related to the signature of the handler
         *  function. For instance, in Java/Swing, an ActionListener would
         *  correspond to a different handler type than a KeyListener because a
         *  keypress event must carry data (which key was pressed) along with
         *  it, whereas an action event need only convey "it happened" and so
         *  has no such payload. */
        public final String handlerType;
        
        /** The identifier of the handler. This will be the name of the method
         *  in the generated controller interface. */
        public final String id;
        
        /** Construct a handler declaration from its type and id.
         *  @param handlerType The type of the handler.
         *  @param id The identifier for the handler. */
        public EventHandler(String handlerType, String id) {
            this.handlerType = handlerType;
            this.id = id;
        }
        
        public String toString() {
            return String.format("Handler: %s \u201c%s\u201d\n", handlerType, id);
        }
    }
    
    /** The filename of the original source file. */
    public final String filename;
    /** A list of all the widgets defined in the file. Immutable. */
    public final List<? extends Widget> widgets;
    /** The set of flags that are set in the file. May be changed if needed,
     *  as by command-line arguments. */
    public final Set<Flag> flags;
    /** A list of the handlers declared in the file. */
    public final List<? extends EventHandler> handlers;
    
    /**
     * @param filename The name of the original file.
     * @param widgets The list to be used for {@link #widgets}.
     * @param flags The set to be used for {@link #flags}.
     * @param handlers The list to be used for {@link #handlers}.
     */
    protected GUIFile(String filename, List<? extends Widget> widgets,
            Set<Flag> flags, List<? extends EventHandler> handlers) {
        this.widgets = widgets;
        this.flags = flags;
        this.handlers = handlers;
        this.filename = filename;
    }
    
    public String toString() {
        final StringBuilder builder = new StringBuilder();
        
        builder.append(String.format("GUI File: %s\n", filename))
            .append(String.format("Project Name: %s\n", getProjectName()));
        
        final String packageId = getPackageId();
        builder.append(String.format("Package: %s\n",
                packageId != null ? packageId : "(default)"))
            .append(String.format("Controller: %s\n", getControllerId()))
            .append(String.format("Flags: %s\n", flags));
            
        if (!handlers.isEmpty()) {
            builder.append(String.format("Event handlers:\n"));
            for (EventHandler handler : handlers)
                builder.append(Output.indent(handler.toString()));
        }
        builder.append("Widgets:\n");
        for (Widget widget : widgets)
            builder.append(Output.indent(widget.toString()));
        
        return builder.toString();
    }

    /**
     * @return The identifier of the controller class to be generated.
     */
    abstract public String getControllerId();

    /**
     * Set the identifier for the controller class to be generated.
     * @param controllerId The new identifier.
     */
    abstract protected void setControllerId(String controllerId);

    /**
     * @return The package identifier for generated files.
     */
    abstract public String getPackageId();

    /**
     * Set the package identifier for generated files.
     * @param packageId The new package identifier.
     */
    abstract public void setPackageId(String packageId);

    /**
     * @return The identifier of the generated class.
     */
    abstract public String getProjectName();

    /**
     * Set the identifier of the generated class.
     * @param projectName The new identifier.
     */
    abstract public void setProjectName(String projectName);
}
