package util;

import java.lang.annotation.*;
import java.util.*;
import java.util.regex.*;

import static java.lang.annotation.ElementType.*;
import static java.lang.annotation.RetentionPolicy.*;

import util.CommandLineParser.OptionSpec;

/**
 * Parses the command line arguments and options passed to a program. The
 * specifications for the options are given in an enum that implements
 * {@link CommandLineParser.OptionSpec} at construction; the parser can then
 * be passed the array of arguments (typically the argument to the main()
 * method), and it will sort out which of them are options to the program,
 * perform actions for each, (see 
 * {@link CommandLineParser.OptionSpec#execute(String, Object)}), and return
 * the other arguments.
 * @param <Opt> The enum specifying the command-line options accepted. Must
 *      implement {@link CommandLineParser.OptionSpec}, and each field must
 *      have a {@link CommandLineParser.OptionSpec.Option} attribute.
 * @param <Cxt> The type of the context in which the members of {@code Opt}
 *      run the {@link CommandLineParser.OptionSpec#execute(String, Object)}
 *      handler method.
 */
public final class CommandLineParser
        <Opt extends Enum<Opt> & OptionSpec<? super Cxt>, Cxt> {
    
    
    /**
     * The interface which the enum passed to 
     * {@link CommandLineParser#buildParser(Class)} must implement. Enum
     * constants implementing this interface implement the 
     * {@link #execute(String, Object)} method, which is called when the
     * option is found in the command line. Also, they carry 
     * {@link OptionSpec.Option} attributes to specify the name of the argument
     * (such as "help" or "verbose") and its abbreviation (such as 'h' or 'v'),
     * as well as other data.
     * @param <Cxt> The type of the context object passed to 
     *      {@link #execute(String, Object)}.
     */
    public interface OptionSpec<Cxt> {
        /**
         * The specification for the argument of an option. There are three
         * kinds of objects: those that never take an argument, those that
         * always do, and those that can take an optional argument.
         */
        enum ArgSpec { 
            /** Specifies that the option takes no argument. */
            NONE,
            /** Specifies that the option must take an argument. */
            REQUIRED,
            /** 
             * Specifies that the option takes an argument, which has a
             * default value.
             */
            OPTIONAL
        }

        /**
         * The annotation giving syntactic specifications for a command-line
         * option. Each enum constant in an OptionSpec enum must carry one.
         * <p>
         * For example, for a program whose --verbose option takes an optional
         * argument specifying a level of verbosity:
         * <p><blockquote>
         * <pre>@Option(name={"verbose", "verbosity"}, abbr='v',
         *        desc="increase level of output (to LEVEL if given)",
         *        arg=ArgSpec.OPTIONAL, argName="LEVEL")
         * VERBOSITY { ... }</pre>
         * </blockquote>
         */
        @Target(FIELD) @Retention(RUNTIME) @Documented
        @interface Option {
            /**
             * The full, long-form name of the option. Excludes the leading
             * "--" used in the command line.
             * <p>
             * An array can be given, specifying multiple synonymous names. The
             * first name given will be used in error message output.
             * <p>
             * The default is to have no long-form name; such an option can
             * only be given by its short form.
             */
            String[] name() default {};
            
            /**
             * The one-character abbreviation for the option. Excludes the
             * leading dash used in the command line.
             * <p>
             * An array can be given, specifying multiple synonymous
             * abbreviations.
             * <p>
             * The default is to have no abbreviation; such an option can
             * only be given by its full form.
             */
            char[] abbr() default {};
            
            /**
             * A brief description of the option, to be included in the
             * "--help" output.
             */
            String desc() default "";
            
            /**
             * An enum constant specifying whether the option takes an
             * argument, and whether that argument is optional.
             */
            ArgSpec arg() default ArgSpec.NONE;
            
            /**
             * The name of the argument, as it will be shown in the "--help"
             * output for the option.
             */
            String argName() default "ARG";
        }
        
        /**
         * What to do with the option when encountered.
         * @param arg The argument given to the argument, or {@code null} if
         *      the argument was given no arguments. (An option annotated with
         *      {@code arg = ArgSpec.REQUIRED} will never be passed
         *      {@code null} for this argument.)
         * @param context The context in which to perform the action. The
         *      exact meaning of this depends on the particular implementor; a
         *      typical context object would simply have a bunch of flags
         *      corresponding to the options that may be set. However, there is
         *      no general requirement on the context object (or the {@code
         *      Cxt} type parameter to OptionSpec); in fact, an implementation
         *      could well have {@code null} for the context and process the
         *      options using {@link CommandLineParser#getOpts()}.
         * @throws OptionException If the argument is somehow invalid. The
         *      meaning of "invalid" is dependent entirely on the
         *      implementation; a typical case would be that an integer is
         *      expected but the given argument doesn't parse as one. These
         *      exceptions are caught by 
         *      {@link CommandLineParser#parse(Object, String[])} and rethrown
         *      as {@link CommandLineParser.InvalidCommandLineException}
         *      objects, after being formatted with the name of the option and
         *      the value of its argument. (Consequently, the message passed to
         *      the OptionException constructor needn't include either.)
         */
        void execute(String arg, Cxt context) throws OptionException;
        
        /**
         * Thrown by the {@link OptionSpec#execute(String, Object)} method of
         * {@link OptionSpec} implementors to indicate an error in the command
         * line. These exceptions are then prettied up by
         * {@link CommandLineParser#parse(Object, String[])} and rethrown. 
         */
        public class OptionException extends Exception {
            protected OptionException() {
                super("");
            }

            protected OptionException(String message, Throwable cause) {
                super(message, cause);
            }

            protected OptionException(String message) {
                super(message);
            }

            protected OptionException(Throwable cause) {
                super("", cause);
            }
        }
    }
    
    /**
     * Annotates an {@link OptionSpec} enum with the usage message to be
     * displayed by the "--help" output or when an error occurs.
     * <p>
     * This string, if given, will be the value of
     * {@link CommandLineParser#getUsage()}.
     */
    @Target(TYPE) @Retention(RUNTIME) @Documented
    public @interface Usage {
        String value();
    }
    
    private class OptData {
        private final Opt opt;
        private final String[] names;
        private final char[] abbrs;
        private final String desc;
        private final OptionSpec.ArgSpec arg;
        private final String argName;
        
        private OptData(Opt opt) {
            this.opt = opt;
            
            OptionSpec.Option annot;
            try {
                annot = specClass.getDeclaredField(opt.name())
                        .getAnnotation(OptionSpec.Option.class);
            } catch (NoSuchFieldException e) {
                // This shouldn't be possible; opt is an enum constant
                // declared in specClass, so there *must* be a field by its
                // name (namely, opt itself).
                throw new IllegalArgumentException(e);
            }
            
            if (annot == null)
                throw new IllegalArgumentException(
                        "option specification missing @Option annotation");
            
            names = annot.name();
            abbrs = annot.abbr();
            desc = annot.desc();
            arg = annot.arg();
            argName = annot.argName();
        }
    }

    /**
     * Exception thrown by {@link CommandLineParser#parse(Object, String[])} on
     * invalid command-line input.
     * <p>
     * Objects of this class are only constructed directly by parse() itself;
     * an {@link OptionSpec} enum throws {@link OptionSpec.OptionException}
     * exceptions, which are then reformatted by parse().
     */
    public static class InvalidCommandLineException extends Exception {
        /** 
         * The option in which the error occurred, or {@code null} if the error
         * was associated with no particular option.
         */
        public final Enum opt;
        /**
         * The argument to the erroneous option, or {@code null} if there was
         * no such.
         */ 
        public final String arg;
        
        private <Opt extends Enum<Opt> & OptionSpec<? super Cxt>, Cxt> 
                InvalidCommandLineException(
                        String msg, Opt opt, String arg,
                        CommandLineParser<Opt, Cxt> me) {
            super(me.formatMsg(msg, opt, arg));
            this.opt = opt;
            this.arg = arg;
        }
        
        @SuppressWarnings("unused")
        private <Opt extends Enum<Opt> & OptionSpec<? super Cxt>, Cxt> 
        InvalidCommandLineException(
                String msg, Opt opt, CommandLineParser<Opt, Cxt> me) {
            this(msg, opt, null, me);
        }
        
        @SuppressWarnings("unused")
        private <Opt extends Enum<Opt> & OptionSpec<? super Cxt>, Cxt> 
        InvalidCommandLineException(
                String msg, String arg, CommandLineParser<Opt, Cxt> me) {
            this(msg, null, arg, me);
        }
        
        @SuppressWarnings("unused")
        private <Opt extends Enum<Opt> & OptionSpec<? super Cxt>, Cxt> 
        InvalidCommandLineException(
                String msg, CommandLineParser<Opt, Cxt> me) {
            this(msg, null, null, me);
        }
    }
    
    private String formatMsg(String msg, Opt opt, String arg) {
        final OptData data = optDataMap.get(opt);
        
        if (opt != null) {
            final String formattedName;
            
            if (data.names.length != 0)
                formattedName = "--" + data.names[0];
            else if (data.abbrs.length != 0)
                formattedName = "-" + data.abbrs[0];
            else
                formattedName = opt.name();

            if (arg != null)
                return String.format("%s: %s: %s", 
                        formattedName, arg, msg);
            else
                return String.format("%s: %s",
                        formattedName, msg);
        } else
            if (arg != null)
                return String.format("%s: %s", arg, msg);
            else 
                return msg;
    }
    
    private final Class<Opt> specClass;
    private final Opt[] optSpecs;
    private final Map<Opt, OptData> optDataMap;
    private final Map<String, OptData> nameMap;
    private final Map<Character, OptData> abbrMap;
    private final String usage;
    private String[] lastArgs = null;
    private Pair<Opt, String>[] lastOpts = null;
    
    private CommandLineParser(Class<Opt> specClass) {
        this.specClass = specClass;
        optSpecs = specClass.getEnumConstants();
        
        final Usage usageAnnot = specClass.getAnnotation(Usage.class);
        usage = usageAnnot != null ? usageAnnot.value() :
                specClass.getName() + " [opts] [args]";
        
        optDataMap = new EnumMap<Opt, OptData>(specClass);
        nameMap = new TreeMap<String, OptData>();
        abbrMap = new TreeMap<Character, OptData>();
        
        for (Opt opt : optSpecs) {
            final OptData data = new OptData(opt);
            optDataMap.put(opt, data);
            for (String name : data.names)
                nameMap.put(name, data);
            for (char abbr : data.abbrs)
                abbrMap.put(abbr, data);
        }
    }
    
    /**
     * Retrieve the parsed options, as a map from the option enum constants to
     * the arguments given. If the order of the options is important, use
     * {@link #getOpts()} instead.
     * @return A map of the options found by the last call to 
     *      {@link #parse(Object, String[])} to an array of the arguments
     *      passed to the option (or {@code null} for options without
     *      arguments) each time it occurred.
     */
    public Map<Opt, String[]> getOptMap() {
        TreeMap<Opt, ArrayList<String>> optMap =
            new TreeMap<Opt, ArrayList<String>>();
        
        for (Pair<Opt, String> opt_arg : getOpts())
            if (optMap.containsKey(opt_arg.a))
                optMap.get(opt_arg.a).add(opt_arg.b);
            else {
                final ArrayList<String> newList = 
                    new ArrayList<String>();
                newList.add(opt_arg.b);
                optMap.put(opt_arg.a, newList);
            }
        
        TreeMap<Opt, String[]> ans =
            new TreeMap<Opt, String[]>();
        
        for (Map.Entry<Opt, ArrayList<String>> entry :
                optMap.entrySet())
            ans.put(entry.getKey(), 
                    entry.getValue().toArray(new String[0]));
        
        return ans;
    }
    
    /**
     * Format a help message, suitable for output by a "--help" option.
     * Includes the usage line (see {@link #getUsage()}), followed by a list of
     * all options with their names, abbreviations, and descriptions.
     * @return The help message.
     */
    // TODO This needs a summary line after the usage line. Perhaps we need
    // a @Summary annotation as well?
    // TODO Word wrapping should happen. But how to know how wide? Should we
    // default to the value of $COLUMNS in the environment, assuming that we
    // intend to output to a terminal (as will usually be the case)?
    public String getHelp() {
        final StringBuilder builder = new StringBuilder(
                getUsage() + "\n\n");
        
        final String[] leftColumn = new String[optDataMap.size()];
        final int leftWidth, totalWidth;
        
        {
            int max = 0, total = 0;
            for (int ix : Functional.range(optSpecs.length)) {
                final Opt opt = optSpecs[ix];
                final OptData data = optDataMap.get(opt);
                
                final String[] forms =
                    new String[data.abbrs.length + data.names.length];
                
                int fIx = 0;
                if (data.names.length == 0)
                    for (char abbr : data.abbrs)
                        switch (data.arg) {
                            case NONE:
                                forms[fIx++] = "-" + abbr;
                                break;
                            case REQUIRED:
                                forms[fIx++] = String.format(
                                        "-%c%s", abbr, data.argName);
                                break;
                            case OPTIONAL:
                                forms[fIx++] = String.format(
                                        "-%c[%s]", abbr, data.argName);
                                break;
                        }
                else
                    for (char abbr : data.abbrs)
                        forms[fIx++] = "-" + abbr;
                
                for (String name : data.names)
                    switch (data.arg) {
                        case NONE:
                            forms[fIx++] = "--" + name;
                            break;
                        case REQUIRED:
                            forms[fIx++] = String.format(
                                    "--%s=%s", name, data.argName);
                            break;
                        case OPTIONAL:
                            forms[fIx++] = String.format(
                                    "--%s[=%s]", name, data.argName);
                    }
                final String itemString = leftColumn[ix] =
                    Functional.join(forms, ", ");
                final int itemLen = itemString.length();
                
                total += itemLen;
                if (itemLen > max)
                    max = itemLen;
            }
            totalWidth = total;
        }
        
        final int lineBreakThreshold = 
            (int) (((double) totalWidth / optSpecs.length) 
                    * LINE_BREAK_THRESHOLD_FACTOR);
        
        {
            int max = 0;
            for (String item : leftColumn)
                if (item.length() > max && item.length() <
                        lineBreakThreshold)
                    max = item.length();
            leftWidth = max;
        }
        
        for (int ix : Functional.range(optSpecs.length)) {
            final Opt opt = optSpecs[ix];
            final OptData data = optDataMap.get(opt);
            
            final String item = leftColumn[ix];
            if (item.length() > leftWidth)
                builder.append(leftColumn[ix] + '\n')
                    .append(String.format(
                        "%" + leftWidth + "s%s%s\n", 
                        "", COLUMN_GUTTER, data.desc
                ));
            else
                builder.append(String.format(
                        "%-" + leftWidth + "s%s%s\n",
                        leftColumn[ix], COLUMN_GUTTER, data.desc
                ));
        }
        return builder.toString();
    }

    private static final float LINE_BREAK_THRESHOLD_FACTOR = 2;
    private static final String COLUMN_GUTTER = "  ";

    /**
     * Look for the plain arguments (i.e. not options) given on the command
     * line, and return them without processing the options.
     * <p>
     * This is something of a hack; it may go away, as I found a better way to
     * do what I wrote this method for.
     * @param cmdLine The command line. This will typically be whatever the
     *      argument to the main() method was.
     * @return The arguments to the program.
     * @throws InvalidCommandLineException If the command line fails to parse;
     *      for instance, if no argument is given to an option that requires
     *      one.
     */
    public String[] extractArguments(String ... cmdLine)
            throws InvalidCommandLineException {
        final ArrayList<String> args = 
            new ArrayList<String>(cmdLine.length);
        
        int ix = -1;
        Opt optPendingArg = null;
        
        while (++ix < cmdLine.length) {
            final String arg = cmdLine[ix];
            if (optPendingArg != null) {
                if (arg.startsWith("-") && !arg.equals("-"))
                    // This will toss an exception
                    break;
                optPendingArg = null;
                continue;
            }
            if (arg.startsWith("-") && !arg.equals("-") && !arg.equals("--")) {
                if (arg.startsWith("--")) {
                    final String nameValuePair = arg.substring(2);
                    Matcher matcher = 
                        NAME_VALUE_REGEX.matcher(nameValuePair);
                    matcher.matches();
                    final String name = matcher.group(1);
                    final String value = matcher.group(2);
                    final OptData optData = nameMap.get(name);
                    switch (optData.arg) {
                        case NONE:
                            if (value != null)
                                throw new InvalidCommandLineException(
                                        "option does not take an argument",
                                        optData.opt, this);
                            // Fall through
                        case OPTIONAL:
                            break;
                        case REQUIRED:
                            if (value == null)
                                optPendingArg = optData.opt;
                            break;
                    }
                } else {
                    int cix;
                    FOR: for (cix = 1; cix < arg.length(); cix++) {
                        final char abbr = arg.charAt(cix);
                        final OptData optData = abbrMap.get(abbr);
                        SWITCH: switch (optData.arg) {
                            case NONE:
                                break SWITCH;
                            case OPTIONAL:
                                break FOR;
                            case REQUIRED:
                                if (cix >= arg.length() - 1)
                                    optPendingArg = optData.opt;
                                break FOR;
                        }
                    }
                }
            } else
                args.add(arg);
        }
        if (optPendingArg != null)
            throw new InvalidCommandLineException("argument required",
                optPendingArg, this);
        
        return args.toArray(new String[0]);
    }
    
    /**
     * Process the options in the command line, and return the other arguments.
     * @param context The context to pass to the 
     *      {@link OptionSpec#execute(String, Object)} method of the OptionSpec
     *      enum for this parser.
     * @param cmdLine The command line. This will typically be whatever the
     *      argument to the main() method was.
     * @return The non-option arguments. The options can later be retrieved by
     *      the {@link #getOpts()} or {@link #getOptMap()} method if necessary,
     *      and this return value will also be available as {@link #getArgs()}.
     * @throws InvalidCommandLineException If there is an error in the command
     *      line.
     */
    public String[] parse(Cxt context, String ... cmdLine) throws 
            InvalidCommandLineException {
        final ArrayList<String> args = 
            new ArrayList<String>(cmdLine.length);
        
        final ArrayList<Pair<Opt, String>> opts =
            new ArrayList<Pair<Opt, String>>(cmdLine.length);
        
        int ix = -1;
        Opt optPendingArg = null;
        while (++ix < cmdLine.length) {
            final String arg = cmdLine[ix];
            if (optPendingArg != null) {
                if (arg.startsWith("-") && !arg.equals("-"))
                    // This will toss an exception
                    break;
                executeOption(optPendingArg, arg, context, opts);
                optPendingArg = null;
                continue;
            }
            if (arg.startsWith("-") && !arg.equals("-") && !arg.equals("--")) {
                if (arg.startsWith("--")) {
                    final String nameValuePair = arg.substring(2);
                    Matcher matcher = 
                        NAME_VALUE_REGEX.matcher(nameValuePair);
                    matcher.matches();
                    final String name = matcher.group(1);
                    final String value = matcher.group(2);
                    final OptData optData = nameMap.get(name);
                    
                    if (optData == null)
                        throw new InvalidCommandLineException(
                                name + ": option not recognized", this);
                    
                    switch (optData.arg) {
                        case NONE:
                            if (value != null)
                                throw new InvalidCommandLineException(
                                        "option does not take an argument",
                                        optData.opt, this);
                            // Fall through
                        case OPTIONAL:
                            executeOption(optData.opt, null, context, opts);
                            break;
                        case REQUIRED:
                            if (value == null)
                                optPendingArg = optData.opt;
                            else
                                executeOption(
                                        optData.opt, value, context, opts);
                            break;
                    }
                } else {
                    int cix;
                    FOR: for (cix = 1; cix < arg.length(); cix++) {
                        final char abbr = arg.charAt(cix);
                        final OptData optData = abbrMap.get(abbr);
                        SWITCH: switch (optData.arg) {
                            case NONE:
                                executeOption(
                                        optData.opt, null, context, opts); 
                                break SWITCH;
                            case OPTIONAL:
                                final String optArg;
                                if (cix < arg.length() - 1)
                                    optArg = arg.substring(cix + 1);
                                else
                                    optArg = null;
                                executeOption(
                                        optData.opt, optArg, context, opts);
                                break FOR;
                            case REQUIRED:
                                if (cix < arg.length() - 1)
                                    executeOption(optData.opt,
                                            arg.substring(cix + 1),
                                            context, opts);
                                else
                                    optPendingArg = optData.opt;
                                break FOR;
                        }
                    }
                }
            } else
                args.add(arg);
        }
        if (optPendingArg != null)
            throw new InvalidCommandLineException("argument required",
                optPendingArg, this);
        
        // yuck ... Java needs arrays of generic types ...
        lastOpts = opts.toArray(this.<Opt, String> makePairArray());
        return (lastArgs = args.toArray(new String[0]));
    }

    private void executeOption(Opt opt, String arg, Cxt context,
                Collection<? super Pair<Opt, String>> acc) 
            throws InvalidCommandLineException {
        try {
            opt.execute(arg, context);
        } catch (OptionSpec.OptionException e) {
            final String msg = e.getMessage();
            throw new InvalidCommandLineException(
                    msg.equals("") ? null : msg, opt, arg, this);
        }
        acc.add(Pair.of(opt, arg));        
    }
    
    private static final Pattern NAME_VALUE_REGEX =
        Pattern.compile("([^=]*)(?:=(.*))?");
    
    // This method's purpose in life is to be unchecked. Yeccch.
    @SuppressWarnings("unchecked")
    private <A extends Object, B extends Object>
            Pair<A, B>[] makePairArray() {
        return (Pair<A, B>[]) new Pair[0];
    }
    
    /**
     * Retrieve the arguments found by the last call to 
     * {@link #parse(Object, String[])}.
     * @return The arguments found by the last call to parse(), or {@code null}
     *      if parse() has not yet been called.
     */
    public String[] getArgs() {
        return lastArgs;
    }

    /**
     * Retrieve the options found by the last call to 
     * {@link #parse(Object, String[])}.
     * @return An array of (option, argument) pairs corresponding to the
     *      options found by the last call to parse() and their arguments, or
     *      {@code null} if parse() has not yet been called.
     */
    public Pair<Opt, String>[] getOpts() {
        return lastOpts;
    }

    /**
     * Get the usage line for the given {@link OptionSpec} enum, or the empty
     * string if none was specified.
     * @return The usage line.
     * @see Usage
     */
    public String getUsage() {
        return usage;
    }

    /**
     * Construct a parser for the given command-line option specifications.
     * @param <Opt> The {@link OptionSpec}-implementing enum type specifying
     *      the command-line options recognized.
     * @param <Cxt> The context type for {@code Opt}.
     * @param specClass The Class object for {@code Opt}.
     * @return A CommandLineParser for the given specifications.
     */
    public static <Opt extends Enum<Opt> & OptionSpec<? super Cxt>, Cxt>
            CommandLineParser<Opt, Cxt>
            buildParser(final Class<Opt> specClass) {
        
        return new CommandLineParser<Opt, Cxt>(specClass);
    }

}