package util;

import java.io.*;

/** A pair of objects of some specified type. 
 *  
 *  Convenient for return values, etc.
 *  
 *  @param <A> The type of the first element of a pair. 
 *  @param <B> The type of the second element of a pair. */
public class Pair<A, B> implements Serializable {
    /** The first element of the pair. */
    public final A a;
    /** The second element of the pair. */
    public final B b;

    /** @param a The first element of the pair.
     *  @param b The second element of the pair. */
    public Pair(A a, B b) {
        this.a = a;
        this.b = b;
    }

    /** A convenience method so that the types of the elements don't have to
     *  be given explicitly.
     *  
     *  <p>Without this method, this would be required to create a Pair:
     *  
     *<p><blockquote><pre>Pair&lt;Integer, String> stuff =
     *    new Pair&lt;Integer, String>(42, "spam");</pre></blockquote><p>
     *      
     *  With this method:
     *  
     *<p><blockquote><pre>Pair&lt;Integer, String> stuff =
     *    Pair.pair(42, "spam");</pre></blockquote>
     *  @param <A> The type of the first element. 
     *  @param <B> The type of the second element.
     *  @param a The first element of the pair.
     *  @param b The second element of the pair.
     *  @return A newly constructed pair. */
    public static <A, B> Pair<A, B> pair(A a, B b) {
        return new Pair<A, B>(a, b);
    }
    
    /** A more readable synonym for {@link #pair}.
     *  @param <A> The type of the first element. 
     *  @param <B> The type of the second element.
     *  @param a The first element of the pair.
     *  @param b The second element of the pair.
     *  @return A newly constructed pair. */
    public static <A, B> Pair<A, B> of(A a, B b) {
        return pair(a, b);
    }
    
    public boolean equals(Object other) {
        if (other instanceof Pair<?, ?>) {
            final Pair<?, ?> otherPair = (Pair<?, ?>) other;
            return a.equals(otherPair.a) && b.equals(otherPair.b);
        }
        return super.equals(other);
    }

    static int rotL(int x, int n) {
        return x << n + x >> (32 - n);
    }
    
    public int hashCode() {
        return rotL(a == null ? 0 : a.hashCode(), 8) ^
            rotL(b == null ? 0 : b.hashCode(), 16);
    }
    
    public String toString() {
        return String.format("(%s, %s)", String.valueOf(a), String.valueOf(b));
    }
}
